# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import threading
from urllib.request import HTTPError, urlopen

import daphne.server
import pytest
from testing_support.fixtures import (
    override_application_settings,
    raise_background_exceptions,
    wait_for_background_threads,
)
from testing_support.sample_asgi_applications import (
    AppWithCall,
    AppWithCallRaw,
    simple_app_v2_raw,
    simple_app_v3,
)
from testing_support.util import get_open_port
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.common.object_names import callable_name

DAPHNE_VERSION = tuple(int(v) for v in daphne.__version__.split(".")[:2])
skip_asgi_3_unsupported = pytest.mark.skipif(DAPHNE_VERSION < (3, 0), reason="ASGI3 unsupported")
skip_asgi_2_unsupported = pytest.mark.skipif(DAPHNE_VERSION >= (3, 0), reason="ASGI2 unsupported")


@pytest.fixture(
    params=(
        pytest.param(
            simple_app_v2_raw,
            marks=skip_asgi_2_unsupported,
        ),
        pytest.param(
            simple_app_v3,
            marks=skip_asgi_3_unsupported,
        ),
        pytest.param(
            AppWithCallRaw(),
            marks=skip_asgi_3_unsupported,
        ),
        pytest.param(
            AppWithCall(),
            marks=skip_asgi_3_unsupported,
        ),
    ),
    ids=("raw", "wrapped", "class_with_call", "class_with_call_double_wrapped"),
)
def app(request, server_and_port):
    app = request.param
    server, _ = server_and_port
    server.application = app
    return app


@pytest.fixture(scope="session")
def port(server_and_port):
    _, port = server_and_port
    return port


@pytest.fixture(scope="session")
def server_and_port():
    port = get_open_port()

    servers = []
    loops = []
    ready = threading.Event()

    def server_run():
        def on_ready():
            if not ready.is_set():
                loops.append(asyncio.get_event_loop())
                servers.append(server)
                ready.set()

        async def fake_app(*args, **kwargs):
            raise RuntimeError("Failed to swap out app.")

        server = daphne.server.Server(
            fake_app,
            endpoints=["tcp:%d:interface=127.0.0.1" % port],
            ready_callable=on_ready,
            signal_handlers=False,
            verbosity=9,
        )

        server.run()

    thread = threading.Thread(target=server_run, daemon=True)
    thread.start()
    assert ready.wait(timeout=10)
    yield servers[0], port

    reactor = daphne.server.reactor
    _ = [loop.call_soon_threadsafe(reactor.stop) for loop in loops]  # Stop all loops
    thread.join(timeout=10)

    if thread.is_alive():
        raise RuntimeError("Thread failed to exit in time.")


@override_application_settings({"transaction_name.naming_scheme": "framework"})
def test_daphne_200(port, app):
    @validate_transaction_metrics(
        callable_name(app),
        custom_metrics=[
            ("Python/Dispatcher/Daphne/%s" % daphne.__version__, 1),
        ],
    )
    @raise_background_exceptions()
    @wait_for_background_threads()
    def response():
        return urlopen("http://localhost:%d" % port, timeout=10)  # nosec

    assert response().status == 200


@override_application_settings({"transaction_name.naming_scheme": "framework"})
@validate_transaction_errors(["builtins:ValueError"])
def test_daphne_500(port, app):
    @validate_transaction_metrics(callable_name(app))
    @raise_background_exceptions()
    @wait_for_background_threads()
    def _test():
        try:
            urlopen("http://localhost:%d/exc" % port)  # nosec
        except HTTPError:
            pass

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from sklearn import __init__  # noqa: needed for get_package_version
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version
from newrelic.packages import six

SKLEARN_VERSION = tuple(map(int, get_package_version("sklearn").split(".")))


@pytest.mark.parametrize(
    "dummy_model_name",
    [
        "DummyClassifier",
        "DummyRegressor",
    ],
)
def test_model_methods_wrapped_in_function_trace(dummy_model_name, run_dummy_model):
    expected_scoped_metrics = {
        "DummyClassifier": [
            ("Function/MLModel/Sklearn/Named/DummyClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/DummyClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/DummyClassifier.predict_log_proba", 1),
            ("Function/MLModel/Sklearn/Named/DummyClassifier.predict_proba", 2 if SKLEARN_VERSION > (1, 0, 0) else 4),
            ("Function/MLModel/Sklearn/Named/DummyClassifier.score", 1),
        ],
        "DummyRegressor": [
            ("Function/MLModel/Sklearn/Named/DummyRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/DummyRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/DummyRegressor.score", 1),
        ],
    }

    expected_transaction_name = (
        "test_dummy_models:test_model_methods_wrapped_in_function_trace.<locals>._test"
        if six.PY3
        else "test_dummy_models:_test"
    )

    @validate_transaction_metrics(
        expected_transaction_name,
        scoped_metrics=expected_scoped_metrics[dummy_model_name],
        rollup_metrics=expected_scoped_metrics[dummy_model_name],
        background_task=True,
    )
    @background_task()
    def _test():
        run_dummy_model(dummy_model_name)

    _test()


@pytest.fixture
def run_dummy_model():
    def _run(dummy_model_name):
        import sklearn.dummy
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split

        X, y = load_iris(return_X_y=True)
        x_train, x_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=0)

        clf = getattr(sklearn.dummy, dummy_model_name)()

        model = clf.fit(x_train, y_train)
        if hasattr(model, "predict"):
            model.predict(x_test)
        if hasattr(model, "score"):
            model.score(x_test, y_test)
        if hasattr(model, "predict_log_proba"):
            model.predict_log_proba(x_test)
        if hasattr(model, "predict_proba"):
            model.predict_proba(x_test)

        return model

    return _run



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time

import pytest
import webtest
from testing_support.fixtures import (
    override_application_settings,
    override_generic_settings,
    reset_core_stats_engine,
    validate_attributes_complete,
    validate_custom_event_count,
    validate_custom_event_in_application_stats_engine,
    validate_request_params_omitted,
)
from testing_support.validators.validate_custom_parameters import (
    validate_custom_parameters,
)
from testing_support.validators.validate_non_transaction_error_event import (
    validate_non_transaction_error_event,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_tt_segment_params import (
    validate_tt_segment_params,
)

from newrelic.api.application import application_instance as application
from newrelic.api.background_task import background_task
from newrelic.api.function_trace import FunctionTrace
from newrelic.api.message_trace import MessageTrace
from newrelic.api.settings import STRIP_EXCEPTION_MESSAGE
from newrelic.api.time_trace import notice_error
from newrelic.api.transaction import (
    add_custom_attribute,
    capture_request_params,
    current_transaction,
    record_custom_event,
)
from newrelic.api.wsgi_application import wsgi_application
from newrelic.common.object_names import callable_name
from newrelic.config import apply_local_high_security_mode_setting
from newrelic.core.agent_protocol import AgentProtocol
from newrelic.core.attribute import (
    DST_ALL,
    DST_ERROR_COLLECTOR,
    DST_TRANSACTION_TRACER,
    Attribute,
)
from newrelic.core.config import Settings, apply_config_setting, global_settings


def test_hsm_configuration_default():
    # Global default should always be off.

    settings = global_settings()
    assert "NEW_RELIC_HIGH_SECURITY_MODE" not in os.environ
    assert settings.high_security is False


_hsm_local_config_file_settings_disabled = [
    {
        "high_security": False,
        "capture_params": True,
        "transaction_tracer.record_sql": "raw",
        "strip_exception_messages.enabled": False,
        "custom_insights_events.enabled": True,
        "ml_insights_events.enabled": True,
        "message_tracer.segment_parameters_enabled": True,
        "application_logging.forwarding.enabled": True,
        "machine_learning.inference_events_value.enabled": True,
    },
    {
        "high_security": False,
        "capture_params": False,
        "transaction_tracer.record_sql": "raw",
        "strip_exception_messages.enabled": False,
        "custom_insights_events.enabled": False,
        "ml_insights_events.enabled": False,
        "message_tracer.segment_parameters_enabled": True,
        "application_logging.forwarding.enabled": True,
        "machine_learning.inference_events_value.enabled": True,
    },
    {
        "high_security": False,
        "capture_params": False,
        "transaction_tracer.record_sql": "obfuscated",
        "strip_exception_messages.enabled": True,
        "custom_insights_events.enabled": True,
        "ml_insights_events.enabled": True,
        "message_tracer.segment_parameters_enabled": False,
        "application_logging.forwarding.enabled": False,
        "machine_learning.inference_events_value.enabled": False,
    },
    {
        "high_security": False,
        "capture_params": False,
        "transaction_tracer.record_sql": "off",
        "strip_exception_messages.enabled": True,
        "custom_insights_events.enabled": False,
        "ml_insights_events.enabled": False,
        "message_tracer.segment_parameters_enabled": False,
        "application_logging.forwarding.enabled": False,
        "machine_learning.inference_events_value.enabled": False,
    },
]

_hsm_local_config_file_settings_enabled = [
    {
        "high_security": True,
        "capture_params": True,
        "transaction_tracer.record_sql": "raw",
        "strip_exception_messages.enabled": True,
        "custom_insights_events.enabled": True,
        "ml_insights_events.enabled": True,
        "message_tracer.segment_parameters_enabled": True,
        "application_logging.forwarding.enabled": False,
        "machine_learning.inference_events_value.enabled": False,
    },
    {
        "high_security": True,
        "capture_params": None,
        "transaction_tracer.record_sql": "raw",
        "strip_exception_messages.enabled": True,
        "custom_insights_events.enabled": True,
        "ml_insights_events.enabled": True,
        "message_tracer.segment_parameters_enabled": True,
        "application_logging.forwarding.enabled": False,
        "machine_learning.inference_events_value.enabled": False,
    },
    {
        "high_security": True,
        "capture_params": True,
        "transaction_tracer.record_sql": "raw",
        "strip_exception_messages.enabled": True,
        "custom_insights_events.enabled": True,
        "ml_insights_events.enabled": True,
        "message_tracer.segment_parameters_enabled": True,
        "application_logging.forwarding.enabled": False,
        "machine_learning.inference_events_value.enabled": False,
    },
    {
        "high_security": True,
        "capture_params": False,
        "transaction_tracer.record_sql": "raw",
        "strip_exception_messages.enabled": True,
        "custom_insights_events.enabled": True,
        "ml_insights_events.enabled": True,
        "message_tracer.segment_parameters_enabled": True,
        "application_logging.forwarding.enabled": True,
        "machine_learning.inference_events_value.enabled": True,
    },
    {
        "high_security": True,
        "capture_params": True,
        "transaction_tracer.record_sql": "obfuscated",
        "strip_exception_messages.enabled": True,
        "custom_insights_events.enabled": True,
        "ml_insights_events.enabled": True,
        "message_tracer.segment_parameters_enabled": True,
        "application_logging.forwarding.enabled": True,
        "machine_learning.inference_events_value.enabled": True,
    },
    {
        "high_security": True,
        "capture_params": True,
        "transaction_tracer.record_sql": "off",
        "strip_exception_messages.enabled": True,
        "custom_insights_events.enabled": True,
        "ml_insights_events.enabled": True,
        "message_tracer.segment_parameters_enabled": False,
        "application_logging.forwarding.enabled": True,
        "machine_learning.inference_events_value.enabled": True,
    },
    {
        "high_security": True,
        "capture_params": True,
        "transaction_tracer.record_sql": "raw",
        "strip_exception_messages.enabled": False,
        "custom_insights_events.enabled": False,
        "ml_insights_events.enabled": False,
        "message_tracer.segment_parameters_enabled": False,
        "application_logging.forwarding.enabled": True,
        "machine_learning.inference_events_value.enabled": True,
    },
]


def parameterize_hsm_local_config(settings_list):
    settings_object_list = []

    for settings in settings_list:
        settings_object = Settings()
        for name, value in settings.items():
            apply_config_setting(settings_object, name, value)
        settings_object_list.append(settings_object)

    return pytest.mark.parametrize("settings", settings_object_list)


@parameterize_hsm_local_config(_hsm_local_config_file_settings_disabled)
def test_local_config_file_override_hsm_disabled(settings):
    original_capture_params = settings.capture_params
    original_record_sql = settings.transaction_tracer.record_sql
    original_strip_messages = settings.strip_exception_messages.enabled
    original_custom_events = settings.custom_insights_events.enabled
    original_ml_events = settings.ml_insights_events.enabled
    original_message_segment_params_enabled = settings.message_tracer.segment_parameters_enabled
    original_application_logging_forwarding_enabled = settings.application_logging.forwarding.enabled
    original_machine_learning_inference_event_value_enabled = settings.machine_learning.inference_events_value.enabled

    apply_local_high_security_mode_setting(settings)

    assert settings.capture_params == original_capture_params
    assert settings.transaction_tracer.record_sql == original_record_sql
    assert settings.strip_exception_messages.enabled == original_strip_messages
    assert settings.custom_insights_events.enabled == original_custom_events
    assert settings.ml_insights_events.enabled == original_ml_events
    assert settings.message_tracer.segment_parameters_enabled == original_message_segment_params_enabled
    assert settings.application_logging.forwarding.enabled == original_application_logging_forwarding_enabled
    assert (
        settings.machine_learning.inference_events_value.enabled
        == original_machine_learning_inference_event_value_enabled
    )


@parameterize_hsm_local_config(_hsm_local_config_file_settings_enabled)
def test_local_config_file_override_hsm_enabled(settings):
    apply_local_high_security_mode_setting(settings)

    assert settings.capture_params not in (True, None)
    assert settings.transaction_tracer.record_sql in ("off", "obfuscated")
    assert settings.strip_exception_messages.enabled
    assert settings.custom_insights_events.enabled is False
    assert settings.ml_insights_events.enabled is False
    assert settings.message_tracer.segment_parameters_enabled is False
    assert settings.application_logging.forwarding.enabled is False
    assert settings.machine_learning.inference_events_value.enabled is False


_server_side_config_settings_hsm_disabled = [
    (
        {
            "high_security": False,
            "capture_params": False,
            "transaction_tracer.record_sql": "obfuscated",
            "strip_exception_messages.enabled": True,
            "custom_insights_events.enabled": False,
            "ml_insights_events.enabled": False,
            "application_logging.forwarding.enabled": False,
            "machine_learning.inference_events_value.enabled": False,
        },
        {
            "agent_config": {
                "capture_params": True,
                "transaction_tracer.record_sql": "raw",
                "strip_exception_messages.enabled": False,
                "custom_insights_events.enabled": True,
                "ml_insights_events.enabled": True,
                "application_logging.forwarding.enabled": True,
                "machine_learning.inference_events_value.enabled": True,
            },
        },
    ),
    (
        {
            "high_security": False,
            "capture_params": True,
            "transaction_tracer.record_sql": "raw",
            "strip_exception_messages.enabled": False,
            "custom_insights_events.enabled": True,
            "ml_insights_events.enabled": True,
            "application_logging.forwarding.enabled": True,
            "machine_learning.inference_events_value.enabled": True,
        },
        {
            "agent_config": {
                "capture_params": False,
                "transaction_tracer.record_sql": "off",
                "strip_exception_messages.enabled": True,
                "custom_insights_events.enabled": False,
                "ml_insights_events.enabled": False,
                "application_logging.forwarding.enabled": False,
                "machine_learning.inference_events_value.enabled": False,
            },
        },
    ),
]

_server_side_config_settings_hsm_enabled = [
    (
        {
            "high_security": True,
            "capture_params": False,
            "transaction_tracer.record_sql": "obfuscated",
            "strip_exception_messages.enabled": True,
            "custom_insights_events.enabled": False,
            "ml_insights_events.enabled": False,
            "application_logging.forwarding.enabled": False,
            "machine_learning.inference_events_value.enabled": False,
        },
        {
            "high_security": True,
            "capture_params": False,
            "transaction_tracer.record_sql": "obfuscated",
            "strip_exception_messages.enabled": True,
            "custom_insights_events.enabled": False,
            "ml_insights_events.enabled": False,
            "application_logging.forwarding.enabled": False,
            "machine_learning.inference_events_value.enabled": False,
            "agent_config": {
                "capture_params": False,
                "transaction_tracer.record_sql": "obfuscated",
                "strip_exception_messages.enabled": True,
                "custom_insights_events.enabled": False,
                "ml_insights_events.enabled": False,
                "application_logging.forwarding.enabled": False,
                "machine_learning.inference_events_value.enabled": False,
            },
        },
    ),
    (
        {
            "high_security": True,
            "capture_params": False,
            "transaction_tracer.record_sql": "obfuscated",
            "strip_exception_messages.enabled": True,
            "custom_insights_events.enabled": False,
            "ml_insights_events.enabled": False,
            "application_logging.forwarding.enabled": False,
            "machine_learning.inference_events_value.enabled": False,
        },
        {
            "high_security": True,
            "capture_params": False,
            "transaction_tracer.record_sql": "obfuscated",
            "strip_exception_messages.enabled": True,
            "custom_insights_events.enabled": False,
            "ml_insights_events.enabled": False,
            "application_logging.forwarding.enabled": False,
            "machine_learning.inference_events_value.enabled": False,
            "agent_config": {
                "capture_params": True,
                "transaction_tracer.record_sql": "raw",
                "strip_exception_messages.enabled": False,
                "custom_insights_events.enabled": True,
                "ml_insights_events.enabled": True,
                "application_logging.forwarding.enabled": True,
                "machine_learning.inference_events_value.enabled": True,
            },
        },
    ),
]


@pytest.mark.parametrize("local_settings,server_settings", _server_side_config_settings_hsm_disabled)
def test_remote_config_fixups_hsm_disabled(local_settings, server_settings):
    assert "high_security" in local_settings
    assert local_settings["high_security"] is False

    assert "high_security" not in server_settings

    agent_config = server_settings["agent_config"]

    original_capture_params = agent_config["capture_params"]
    original_record_sql = agent_config["transaction_tracer.record_sql"]
    original_strip_messages = agent_config["strip_exception_messages.enabled"]
    original_custom_events = agent_config["custom_insights_events.enabled"]
    original_ml_events = agent_config["ml_insights_events.enabled"]
    original_log_forwarding = agent_config["application_logging.forwarding.enabled"]
    original_machine_learning_events = agent_config["machine_learning.inference_events_value.enabled"]

    _settings = global_settings()
    settings = override_generic_settings(_settings, local_settings)(AgentProtocol._apply_high_security_mode_fixups)(
        server_settings, _settings
    )

    agent_config = settings["agent_config"]

    assert "high_security" not in settings

    assert agent_config["capture_params"] == original_capture_params
    assert agent_config["transaction_tracer.record_sql"] == original_record_sql
    assert agent_config["strip_exception_messages.enabled"] == original_strip_messages
    assert agent_config["custom_insights_events.enabled"] == original_custom_events
    assert agent_config["ml_insights_events.enabled"] == original_ml_events
    assert agent_config["application_logging.forwarding.enabled"] == original_log_forwarding
    assert agent_config["machine_learning.inference_events_value.enabled"] == original_machine_learning_events


@pytest.mark.parametrize("local_settings,server_settings", _server_side_config_settings_hsm_enabled)
def test_remote_config_fixups_hsm_enabled(local_settings, server_settings):
    assert "high_security" in local_settings
    assert local_settings["high_security"] is True

    assert "high_security" in server_settings

    _settings = global_settings()
    settings = override_generic_settings(_settings, local_settings)(AgentProtocol._apply_high_security_mode_fixups)(
        server_settings, _settings
    )

    agent_config = settings["agent_config"]

    assert "high_security" not in settings
    assert "capture_params" not in settings
    assert "transaction_tracer.record_sql" not in settings
    assert "strip_exception_messages.enabled" not in settings
    assert "custom_insights_events.enabled" not in settings
    assert "ml_insights_events.enabled" not in settings
    assert "application_logging.forwarding.enabled" not in settings
    assert "machine_learning.inference_events_value.enabled" not in settings

    assert "capture_params" not in agent_config
    assert "transaction_tracer.record_sql" not in agent_config
    assert "strip_exception_messages.enabled" not in agent_config
    assert "custom_insights_events.enabled" not in agent_config
    assert "ml_insights_events.enabled" not in agent_config
    assert "application_logging.forwarding.enabled" not in agent_config
    assert "machine_learning.inference_events_value.enabled" not in agent_config


def test_remote_config_hsm_fixups_server_side_disabled():
    local_settings = {"high_security": True}
    server_settings = {"high_security": True}

    _settings = global_settings()
    settings = override_generic_settings(_settings, local_settings)(AgentProtocol._apply_high_security_mode_fixups)(
        server_settings, _settings
    )

    assert "high_security" not in settings


_test_transaction_settings_hsm_disabled = {"high_security": False}

# Normally, in HSM the exception message would be stripped just by turning on
# high_security. However, these tests (like all of our tests) overrides the
# settings after agent initialization where this setting is fixed up.

_test_transaction_settings_hsm_enabled = {
    "high_security": True,
    "strip_exception_messages.enabled": True,
    "custom_insights_events.enabled": False,
    "ml_insights_events.enabled": False,
}


@override_application_settings(_test_transaction_settings_hsm_disabled)
@validate_custom_parameters(required_params=[("key", "value")])
@background_task()
def test_other_transaction_custom_parameters_hsm_disabled():
    add_custom_attribute("key", "value")


@override_application_settings(_test_transaction_settings_hsm_disabled)
@validate_custom_parameters(required_params=[("key-1", "value-1"), ("key-2", "value-2")])
@background_task()
def test_other_transaction_multiple_custom_parameters_hsm_disabled():
    transaction = current_transaction()
    transaction.add_custom_attributes([("key-1", "value-1"), ("key-2", "value-2")])


@override_application_settings(_test_transaction_settings_hsm_enabled)
@validate_custom_parameters(forgone_params=[("key", "value")])
@background_task()
def test_other_transaction_custom_parameters_hsm_enabled():
    add_custom_attribute("key", "value")


@override_application_settings(_test_transaction_settings_hsm_enabled)
@validate_custom_parameters(forgone_params=[("key-1", "value-1"), ("key-2", "value-2")])
@background_task()
def test_other_transaction_multiple_custom_parameters_hsm_enabled():
    transaction = current_transaction()
    transaction.add_custom_attributes([("key-1", "value-1"), ("key-2", "value-2")])


class TestException(Exception):
    pass


_test_exception_name = "%s:%s" % (__name__, TestException.__name__)


@override_application_settings(_test_transaction_settings_hsm_disabled)
@validate_transaction_errors(errors=[(_test_exception_name, "test message")], required_params=[("key-2", "value-2")])
@validate_custom_parameters(required_params=[("key-1", "value-1")])
@background_task()
def test_other_transaction_error_parameters_hsm_disabled():
    add_custom_attribute("key-1", "value-1")
    try:
        raise TestException("test message")
    except Exception:
        notice_error(attributes={"key-2": "value-2"})


@override_application_settings(_test_transaction_settings_hsm_enabled)
@validate_transaction_errors(
    errors=[(_test_exception_name, STRIP_EXCEPTION_MESSAGE)], forgone_params=[("key-2", "value-2")]
)
@validate_custom_parameters(forgone_params=[("key-1", "value-1")])
@background_task()
def test_other_transaction_error_parameters_hsm_enabled():
    add_custom_attribute("key-1", "value-1")
    try:
        raise TestException("test message")
    except Exception:
        notice_error(attributes={"key-2": "value-2"})


_err_message = "Error! :("
_intrinsic_attributes = {
    "error.class": callable_name(TestException),
    "error.message": _err_message,
    "error.expected": False,
}


@reset_core_stats_engine()
@override_application_settings(_test_transaction_settings_hsm_disabled)
@validate_non_transaction_error_event(required_intrinsics=_intrinsic_attributes, required_user={"key-1": "value-1"})
def test_non_transaction_error_parameters_hsm_disabled():
    try:
        raise TestException(_err_message)
    except Exception:
        app = application()
        notice_error(attributes={"key-1": "value-1"}, application=app)


_intrinsic_attributes = {
    "error.class": callable_name(TestException),
    "error.message": STRIP_EXCEPTION_MESSAGE,
    "error.expected": False,
}


@reset_core_stats_engine()
@override_application_settings(_test_transaction_settings_hsm_enabled)
@validate_non_transaction_error_event(required_intrinsics=_intrinsic_attributes, forgone_user={"key-1": "value-1"})
def test_non_transaction_error_parameters_hsm_enabled():
    try:
        raise TestException(_err_message)
    except Exception:
        app = application()
        notice_error(attributes={"key-1": "value-1"}, application=app)


@wsgi_application()
def target_wsgi_application_capture_params(environ, start_response):
    status = "200 OK"
    output = b"Hello World!"

    response_headers = [("Content-Type", "text/plain; charset=utf-8"), ("Content-Length", str(len(output)))]
    start_response(status, response_headers)

    return [output]


@wsgi_application()
def target_wsgi_application_capture_params_api_called(environ, start_response):
    status = "200 OK"
    output = b"Hello World!"

    capture_request_params(True)
    response_headers = [("Content-Type", "text/plain; charset=utf-8"), ("Content-Length", str(len(output)))]
    start_response(status, response_headers)

    return [output]


_test_transaction_settings_hsm_enabled_capture_params = {"high_security": True, "capture_params": False}


@override_application_settings(_test_transaction_settings_hsm_enabled_capture_params)
@validate_request_params_omitted()
def test_transaction_hsm_enabled_environ_capture_request_params():
    target_application = webtest.TestApp(target_wsgi_application_capture_params)

    target_application.get("/", params="key-1=value-1")


@override_application_settings(_test_transaction_settings_hsm_enabled_capture_params)
@validate_request_params_omitted()
def test_transaction_hsm_enabled_environ_capture_request_params_disabled():
    target_application = webtest.TestApp(target_wsgi_application_capture_params)

    environ = {}
    environ["newrelic.capture_request_params"] = False

    target_application.get("/", params="key-1=value-1", extra_environ=environ)


@override_application_settings(_test_transaction_settings_hsm_enabled_capture_params)
@validate_request_params_omitted()
def test_transaction_hsm_enabled_environ_capture_request_params_enabled():
    target_application = webtest.TestApp(target_wsgi_application_capture_params)

    environ = {}
    environ["newrelic.capture_request_params"] = True

    target_application.get("/", params="key-1=value-1", extra_environ=environ)


@override_application_settings(_test_transaction_settings_hsm_enabled_capture_params)
@validate_request_params_omitted()
def test_transaction_hsm_enabled_environ_capture_request_params_api_called():
    target_application = webtest.TestApp(target_wsgi_application_capture_params_api_called)

    target_application.get("/", params="key-1=value-1")


# Custom events

_event_type = "SimpleAppEvent"
_params = {"snowman": "\u2603", "foo": "bar"}


@wsgi_application()
def simple_custom_event_app(environ, start_response):
    record_custom_event(_event_type, _params)
    start_response(status="200 OK", response_headers=[])
    return []


_intrinsics = {"type": _event_type, "timestamp": time.time()}

_required_event = [_intrinsics, _params]


@reset_core_stats_engine()
@validate_custom_event_in_application_stats_engine(_required_event)
@override_application_settings(_test_transaction_settings_hsm_disabled)
def test_custom_event_hsm_disabled():
    target_application = webtest.TestApp(simple_custom_event_app)
    target_application.get("/")


@reset_core_stats_engine()
@validate_custom_event_count(count=0)
@override_application_settings(_test_transaction_settings_hsm_enabled)
def test_custom_event_hsm_enabled():
    target_application = webtest.TestApp(simple_custom_event_app)
    target_application.get("/")


# Make sure we don't display the query string in 'request.headers.referer'
# Attribute will exist, and value will have query string stripped off.

_required_attr = Attribute(
    name="request.headers.referer",
    value="http://example.com/blah",
    destinations=DST_TRANSACTION_TRACER | DST_ERROR_COLLECTOR,
)

# Check that the unsanitized version isn't present either, for any
# destinations.

_forgone_attr = Attribute(
    name="request.headers.referer", value="http://example.com/blah?query=value", destinations=DST_ALL
)

_required_attrs = [_required_attr]
_foregone_attrs = [_forgone_attr]


@override_application_settings(_test_transaction_settings_hsm_enabled_capture_params)
@validate_attributes_complete("agent", _required_attrs, _foregone_attrs)
def test_http_referrer_url_is_sanitized_in_hsm():
    target_application = webtest.TestApp(target_wsgi_application_capture_params)

    environ = {}
    environ["HTTP_REFERER"] = "http://example.com/blah?query=value"

    target_application.get("/", extra_environ=environ)


@pytest.mark.parametrize("hsm_enabled", [True, False])
def test_function_trace_params_dropped_in_hsm(hsm_enabled):
    @background_task()
    def _test():
        with FunctionTrace("trace", params={"secret": "super secret"}):
            pass

    if hsm_enabled:
        _test = override_application_settings(_test_transaction_settings_hsm_enabled_capture_params)(_test)
        _test = validate_tt_segment_params(forgone_params=("secret",))(_test)
    else:
        _test = override_application_settings(_test_transaction_settings_hsm_disabled)(_test)
        _test = validate_tt_segment_params(present_params=("secret",))(_test)

    _test()


@pytest.mark.parametrize("hsm_enabled", [True, False])
def test_message_trace_params_dropped_in_hsm(hsm_enabled):
    @background_task()
    def _test():
        with MessageTrace("library", "operation", "dest_type", "dest_name", params={"secret": "super secret"}):
            pass

    if hsm_enabled:
        _test = override_application_settings(_test_transaction_settings_hsm_enabled_capture_params)(_test)
        _test = validate_tt_segment_params(forgone_params=("secret",))(_test)
    else:
        _test = override_application_settings(_test_transaction_settings_hsm_disabled)(_test)
        _test = validate_tt_segment_params(present_params=("secret",))(_test)

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from compat import basic_consume
import pika
import six

from newrelic.api.background_task import background_task

from conftest import QUEUE, BODY
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.db_settings import rabbitmq_settings

DB_SETTINGS = rabbitmq_settings()[0]


class CustomPikaChannel(pika.channel.Channel):
    def _on_getok(self, method_frame, header_frame, body):
        if self._on_getok_callback is not None:
            callback = self._on_getok_callback
            self._on_getok_callback = None
            # Use kwargs to pass values to callback
            callback(channel=self,
                    method_frame=method_frame.method,
                    header_frame=header_frame.properties,
                    body=body)

    def _on_deliver(self, method_frame, header_frame, body):
        consumer_tag = method_frame.method.consumer_tag

        if consumer_tag in self._cancelled:
            if self.is_open and consumer_tag not in self._consumers_with_noack:
                self.basic_reject(method_frame.method.delivery_tag)
            return

        self._consumers[consumer_tag](channel=self,
                method_frame=method_frame.method,
                header_frame=header_frame.properties,
                body=body)


class CustomPikaConnection(pika.SelectConnection):
    def _create_channel(self, channel_number, on_open_callback):
        return CustomPikaChannel(self, channel_number, on_open_callback)


_test_select_connection_supportability_metrics = [
    ('Supportability/hooks/pika/kwargs_error', 1)
]


@validate_transaction_metrics(
        ('test_pika_supportability:'
                'test_select_connection_supportability_in_txn'),
        scoped_metrics=(),
        rollup_metrics=_test_select_connection_supportability_metrics,
        background_task=True)
@background_task()
def test_select_connection_supportability_in_txn(producer):
    def on_message(channel, method_frame, header_frame, body):
        assert method_frame
        assert body == BODY
        channel.basic_ack(method_frame.delivery_tag)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    def on_open_channel(channel):
        channel.basic_get(callback=on_message, queue=QUEUE)

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = CustomPikaConnection(
            pika.ConnectionParameters(DB_SETTINGS['host']),
            on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        # Start the IOLoop again so Pika can communicate, it will stop on its
        # own when the connection is closed
        connection.ioloop.stop()
        raise


if six.PY3:
    _txn_name = ('test_pika_supportability:'
            'test_select_connection_supportability_outside_txn.'
            '<locals>.on_message')
else:
    _txn_name = (
        'test_pika_supportability:on_message')


@validate_transaction_metrics(
        _txn_name,
        scoped_metrics=(),
        rollup_metrics=_test_select_connection_supportability_metrics,
        background_task=True,
        group='Message/RabbitMQ/Exchange/Unknown')
def test_select_connection_supportability_outside_txn(producer):
    def on_message(channel, method_frame, header_frame, body):
        assert method_frame
        assert body == BODY
        channel.basic_ack(method_frame.delivery_tag)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    def on_open_channel(channel):
        basic_consume(channel, QUEUE, on_message)

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = CustomPikaConnection(
            pika.ConnectionParameters(DB_SETTINGS['host']),
            on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        # Start the IOLoop again so Pika can communicate, it will stop on its
        # own when the connection is closed
        connection.ioloop.stop()
        raise



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import falcon
import webtest


try:
    from falcon import HTTPRouteNotFound
    NOT_FOUND_ERROR_NAME = 'falcon.errors:HTTPRouteNotFound'
except ImportError:
    NOT_FOUND_ERROR_NAME = 'falcon.errors:HTTPNotFound'


def _bind_response(*args, **kwargs):
    args = list(args)
    args.extend(kwargs.values())
    for arg in args:
        if hasattr(arg, 'status'):
            return arg


class BadGetRequest(ValueError):
    pass


class BadPutRequest(ValueError):
    pass


class Index(object):
    def on_get(self, req, resp):
        """Handles GET requests"""
        resp.content_type = 'application/json'
        resp.data = b'{"status": "ok"}'


class BadResponse(object):
    def on_get(self, req, resp):
        raise BadGetRequest()

    def on_put(self, req, resp):
        raise BadPutRequest()


try:
    application = falcon.App()
    name_prefix = 'falcon.app:App'
except AttributeError:
    application = falcon.API()
    name_prefix = 'falcon.api:API'


def bad_error_handler(*args, **kwargs):
    resp = _bind_response(*args, **kwargs)

    # This endpoint is explicitly not doing the correct thing. Status is
    # expected to be a type str with a status code + explanation. The intent
    # here is to test what happens if we don't parse the response code
    # correctly.
    resp.status = 200


application.add_route('/', Index())
application.add_route('/bad_response', BadResponse())
application.add_error_handler(BadGetRequest, bad_error_handler)

_target_application = webtest.TestApp(application)

# Put exception class here for convenience
_target_application.BadPutRequest = BadPutRequest

# Put names here for convenience
_target_application.name_prefix = name_prefix
_target_application.not_found_error = NOT_FOUND_ERROR_NAME



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import postgresql.driver.dbapi20
from testing_support.db_settings import postgresql_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_database_trace_inputs import (
    validate_database_trace_inputs,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task

DB_SETTINGS = postgresql_settings()[0]

_test_execute_via_cursor_scoped_metrics = [
    ("Function/postgresql.driver.dbapi20:connect", 1),
    ("Function/postgresql.driver.dbapi20:Connection.__enter__", 1),
    ("Function/postgresql.driver.dbapi20:Connection.__exit__", 1),
    ("Datastore/statement/Postgres/%s/select" % DB_SETTINGS["table_name"], 1),
    ("Datastore/statement/Postgres/%s/insert" % DB_SETTINGS["table_name"], 1),
    ("Datastore/statement/Postgres/%s/update" % DB_SETTINGS["table_name"], 1),
    ("Datastore/statement/Postgres/%s/delete" % DB_SETTINGS["table_name"], 1),
    ("Datastore/statement/Postgres/now/call", 1),
    ("Datastore/statement/Postgres/pg_sleep/call", 1),
    ("Datastore/operation/Postgres/drop", 1),
    ("Datastore/operation/Postgres/create", 1),
    ("Datastore/operation/Postgres/commit", 3),
    ("Datastore/operation/Postgres/rollback", 1),
    ("Datastore/operation/Postgres/other", 1),
]

_test_execute_via_cursor_rollup_metrics = [
    ("Datastore/all", 14),
    ("Datastore/allOther", 14),
    ("Datastore/Postgres/all", 14),
    ("Datastore/Postgres/allOther", 14),
    ("Datastore/operation/Postgres/select", 1),
    ("Datastore/statement/Postgres/%s/select" % DB_SETTINGS["table_name"], 1),
    ("Datastore/operation/Postgres/insert", 1),
    ("Datastore/statement/Postgres/%s/insert" % DB_SETTINGS["table_name"], 1),
    ("Datastore/operation/Postgres/update", 1),
    ("Datastore/statement/Postgres/%s/update" % DB_SETTINGS["table_name"], 1),
    ("Datastore/operation/Postgres/delete", 1),
    ("Datastore/statement/Postgres/%s/delete" % DB_SETTINGS["table_name"], 1),
    ("Datastore/operation/Postgres/drop", 1),
    ("Datastore/operation/Postgres/create", 1),
    ("Datastore/statement/Postgres/now/call", 1),
    ("Datastore/statement/Postgres/pg_sleep/call", 1),
    ("Datastore/operation/Postgres/call", 2),
    ("Datastore/operation/Postgres/commit", 3),
    ("Datastore/operation/Postgres/rollback", 1),
    ("Datastore/operation/Postgres/other", 1),
    ("Datastore/instance/Postgres/%s/%s" % (instance_hostname(DB_SETTINGS["host"]), DB_SETTINGS["port"]), 13),
    ("Function/postgresql.driver.dbapi20:connect", 1),
    ("Function/postgresql.driver.dbapi20:Connection.__enter__", 1),
    ("Function/postgresql.driver.dbapi20:Connection.__exit__", 1),
]


@validate_transaction_metrics(
    "test_database:test_execute_via_cursor",
    scoped_metrics=_test_execute_via_cursor_scoped_metrics,
    rollup_metrics=_test_execute_via_cursor_rollup_metrics,
    background_task=True,
)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor():
    with postgresql.driver.dbapi20.connect(
        database=DB_SETTINGS["name"],
        user=DB_SETTINGS["user"],
        password=DB_SETTINGS["password"],
        host=DB_SETTINGS["host"],
        port=DB_SETTINGS["port"],
    ) as connection:
        cursor = connection.cursor()

        cursor.execute("""drop table if exists %s""" % DB_SETTINGS["table_name"])

        cursor.execute("""create table %s """ % DB_SETTINGS["table_name"] + """(a integer, b real, c text)""")

        cursor.executemany(
            """insert into %s """ % DB_SETTINGS["table_name"] + """values (%s, %s, %s)""",
            [(1, 1.0, "1.0"), (2, 2.2, "2.2"), (3, 3.3, "3.3")],
        )

        cursor.execute("""select * from %s""" % DB_SETTINGS["table_name"])

        cursor.execute(
            """with temporaryTable (averageValue) as (select avg(b) from %s) """ % DB_SETTINGS["table_name"]
            + """select * from %s,temporaryTable """ % DB_SETTINGS["table_name"]
            + """where %s.b > temporaryTable.averageValue""" % DB_SETTINGS["table_name"]
        )

        cursor.execute(
            """update %s """ % DB_SETTINGS["table_name"] + """set a=%s, b=%s, c=%s where a=%s""",
            (4, 4.0, "4.0", 1),
        )

        cursor.execute("""delete from %s where a=2""" % DB_SETTINGS["table_name"])

        connection.commit()

        cursor.callproc("now", ())
        cursor.callproc("pg_sleep", (0.25,))

        connection.rollback()
        connection.commit()


_test_rollback_on_exception_scoped_metrics = [
    ("Function/postgresql.driver.dbapi20:connect", 1),
    ("Function/postgresql.driver.dbapi20:Connection.__enter__", 1),
    ("Function/postgresql.driver.dbapi20:Connection.__exit__", 1),
    ("Datastore/operation/Postgres/rollback", 1),
]

_test_rollback_on_exception_rollup_metrics = [
    ("Datastore/all", 2),
    ("Datastore/allOther", 2),
    ("Datastore/Postgres/all", 2),
    ("Datastore/Postgres/allOther", 2),
]


@validate_transaction_metrics(
    "test_database:test_rollback_on_exception",
    scoped_metrics=_test_rollback_on_exception_scoped_metrics,
    rollup_metrics=_test_rollback_on_exception_rollup_metrics,
    background_task=True,
)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_rollback_on_exception():
    try:
        with postgresql.driver.dbapi20.connect(
            database=DB_SETTINGS["name"],
            user=DB_SETTINGS["user"],
            password=DB_SETTINGS["password"],
            host=DB_SETTINGS["host"],
            port=DB_SETTINGS["port"],
        ):
            raise RuntimeError("error")

    except RuntimeError:
        pass



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from newrelic.api.background_task import background_task
from newrelic.core.database_utils import SQLConnections

from testing_support.validators.validate_database_node import validate_database_node
from testing_support.validators.validate_sql_obfuscation import validate_sql_obfuscation
from utils import DB_SETTINGS


@pytest.fixture()
def psycopg2_cursor():
    import psycopg2

    connection = psycopg2.connect(
        database=DB_SETTINGS["name"],
        user=DB_SETTINGS["user"],
        password=DB_SETTINGS["password"],
        host=DB_SETTINGS["host"],
        port=DB_SETTINGS["port"],
    )

    try:
        cursor = connection.cursor()

        cursor.execute("drop table if exists %s" % DB_SETTINGS["table_name"])
        cursor.execute("create table %s (b text, c text)" % DB_SETTINGS["table_name"])

        yield cursor

    finally:
        connection.close()


_quoting_style_tests = [
    (
        "SELECT * FROM %s WHERE b='2'" % DB_SETTINGS["table_name"],
        "SELECT * FROM %s WHERE b=?" % DB_SETTINGS["table_name"],
    ),
    (
        "SELECT * FROM %s WHERE b=$func$2$func$" % DB_SETTINGS["table_name"],
        "SELECT * FROM %s WHERE b=?" % DB_SETTINGS["table_name"],
    ),
    (
        "SELECT * FROM %s WHERE b=U&'2'" % DB_SETTINGS["table_name"],
        "SELECT * FROM %s WHERE b=U&?" % DB_SETTINGS["table_name"],
    ),
]


@pytest.mark.parametrize("sql,obfuscated", _quoting_style_tests)
def test_quoting_styles(psycopg2_cursor, sql, obfuscated):
    @validate_sql_obfuscation([obfuscated])
    @background_task()
    def test():
        psycopg2_cursor.execute(sql)

    test()


_parameter_tests = [
    (
        "SELECT * FROM " + DB_SETTINGS["table_name"] + " where b=%s",
        "SELECT * FROM " + DB_SETTINGS["table_name"] + " where b=%s",
    ),
]


@pytest.mark.parametrize("sql,obfuscated", _parameter_tests)
def test_parameters(psycopg2_cursor, sql, obfuscated):
    @validate_sql_obfuscation([obfuscated])
    @background_task()
    def test():
        psycopg2_cursor.execute(sql, ("hello",))

    test()


def no_explain_plan(node):
    sql_connections = SQLConnections()
    explain_plan = node.explain_plan(sql_connections)
    assert explain_plan is None


def any_length_explain_plan(node):
    if node.statement.operation != "select":
        return

    sql_connections = SQLConnections()
    explain_plan = node.explain_plan(sql_connections)
    assert explain_plan and len(explain_plan) > 0


_test_explain_plans = [
    (
        "SELECT (b, c) FROM  %s ; SELECT (b, c) FROM %s"
        % (DB_SETTINGS["table_name"], DB_SETTINGS["table_name"]),
        no_explain_plan,
    ),
    (
        "SELECT (b, c) FROM  %s ; SELECT (b, c) FROM %s;"
        % (DB_SETTINGS["table_name"], DB_SETTINGS["table_name"]),
        no_explain_plan,
    ),
    ("SELECT (b, c) FROM  %s WHERE b=';'" % DB_SETTINGS["table_name"], no_explain_plan),
    (";SELECT (b, c) FROM %s" % DB_SETTINGS["table_name"], no_explain_plan),
    ("SELECT (b, c) FROM  %s" % DB_SETTINGS["table_name"], any_length_explain_plan),
    ("SELECT (b, c) FROM  %s;" % DB_SETTINGS["table_name"], any_length_explain_plan),
    (
        "SELECT (b, c) FROM  %s;;;;;;" % DB_SETTINGS["table_name"],
        any_length_explain_plan,
    ),
    (
        "SELECT (b, c) FROM  %s;\n\n" % DB_SETTINGS["table_name"],
        any_length_explain_plan,
    ),
]


@pytest.mark.parametrize("sql,validator", _test_explain_plans)
def test_explain_plans(sql, validator):
    @validate_database_node(validator)
    @background_task()
    def test():
        import psycopg2

        connection = psycopg2.connect(
            database=DB_SETTINGS["name"],
            user=DB_SETTINGS["user"],
            password=DB_SETTINGS["password"],
            host=DB_SETTINGS["host"],
            port=DB_SETTINGS["port"],
        )

        try:
            cursor = connection.cursor()
            cursor.execute("drop table if exists %s" % DB_SETTINGS["table_name"])
            cursor.execute(
                "create table %s (b text, c text)" % DB_SETTINGS["table_name"]
            )

            cursor.execute(sql)

        finally:
            connection.commit()
            connection.close()

    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

import numpy as np
import pandas as pd
import pytest
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task

# This will act as the UUID for `prediction_id`
ML_METRIC_FORCED_UUID = "0b59992f-2349-4a46-8de1-696d3fe1088b"


@pytest.fixture(scope="function")
def force_uuid(monkeypatch):
    monkeypatch.setattr(uuid, "uuid4", lambda *a, **k: ML_METRIC_FORCED_UUID)


_test_prediction_stats_tags = frozenset(
    {("modelName", "DummyClassifier"), ("prediction_id", ML_METRIC_FORCED_UUID), ("model_version", "0.0.0")}
)


@pytest.mark.parametrize(
    "x_train,y_train,x_test,metrics",
    [
        (
            [[0, 0], [1, 1]],
            [0, 1],
            [[2.0, 2.0], [0, 0.5]],
            [
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Mean", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Percentile25",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Percentile50",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Percentile75",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Count", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Mean", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Percentile25",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Percentile50",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Percentile75",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Count", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Mean", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile25", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile50", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile75", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Count", _test_prediction_stats_tags, 1),
            ],
        ),
        (
            np.array([[0, 0], [1, 1]]),
            [0, 1],
            np.array([[2.0, 2.0], [0, 0.5]]),
            [
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Mean", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Percentile25",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Percentile50",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Percentile75",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Count", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Mean", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Percentile25",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Percentile50",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Percentile75",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Count", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Mean", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile25", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile50", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile75", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Count", _test_prediction_stats_tags, 1),
            ],
        ),
        (
            np.array([["a", 0, 4], ["b", 1, 3]], dtype="<U4"),
            [0, 1],
            np.array([["a", 2.0, 3], ["b", 0.5, 4]], dtype="<U4"),
            [
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Mean", _test_prediction_stats_tags, None),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Percentile25",
                    _test_prediction_stats_tags,
                    None,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Percentile50",
                    _test_prediction_stats_tags,
                    None,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Percentile75",
                    _test_prediction_stats_tags,
                    None,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/StandardDeviation",
                    _test_prediction_stats_tags,
                    None,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Min", _test_prediction_stats_tags, None),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Max", _test_prediction_stats_tags, None),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/0/Count", _test_prediction_stats_tags, None),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Mean", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Percentile25",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Percentile50",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Percentile75",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/1/Count", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Mean", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile25", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile50", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile75", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Count", _test_prediction_stats_tags, 1),
            ],
        ),
        (
            pd.DataFrame({"col1": [0, 0], "col2": [1, 1]}),
            [0, 1],
            pd.DataFrame({"col1": [2.0, 2.0], "col2": [0, 0.5]}),
            [
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Mean", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Percentile25",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Percentile50",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Percentile75",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Count", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Mean", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Percentile25",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Percentile50",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Percentile75",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Count", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Mean", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile25", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile50", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile75", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Count", _test_prediction_stats_tags, 1),
            ],
        ),
        (
            pd.DataFrame({"col1": ["a", "b"], "col2": [0.5, 1]}, dtype="<U4"),
            [0, 1],
            pd.DataFrame({"col1": ["a", "b"], "col2": [0, 0.5]}, dtype="<U4"),
            [
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Mean", _test_prediction_stats_tags, None),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Percentile25",
                    _test_prediction_stats_tags,
                    None,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Percentile50",
                    _test_prediction_stats_tags,
                    None,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Percentile75",
                    _test_prediction_stats_tags,
                    None,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/StandardDeviation",
                    _test_prediction_stats_tags,
                    None,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Min", _test_prediction_stats_tags, None),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Max", _test_prediction_stats_tags, None),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Count", _test_prediction_stats_tags, None),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Mean", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Percentile25",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Percentile50",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Percentile75",
                    _test_prediction_stats_tags,
                    1,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Count", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Mean", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile25", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile50", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile75", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Count", _test_prediction_stats_tags, 1),
            ],
        ),
        (
            pd.DataFrame({"col1": ["a", "b"], "col2": [True, False]}, dtype="<U4"),
            [0, 1],
            pd.DataFrame({"col1": ["a", "c"], "col2": [True, False]}, dtype="<U4"),
            [
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Mean", _test_prediction_stats_tags, None),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Percentile25",
                    _test_prediction_stats_tags,
                    None,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Percentile50",
                    _test_prediction_stats_tags,
                    None,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Percentile75",
                    _test_prediction_stats_tags,
                    None,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/StandardDeviation",
                    _test_prediction_stats_tags,
                    None,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Min", _test_prediction_stats_tags, None),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Max", _test_prediction_stats_tags, None),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col1/Count", _test_prediction_stats_tags, None),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Mean", _test_prediction_stats_tags, None),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Percentile25",
                    _test_prediction_stats_tags,
                    None,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Percentile50",
                    _test_prediction_stats_tags,
                    None,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Percentile75",
                    _test_prediction_stats_tags,
                    None,
                ),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/StandardDeviation",
                    _test_prediction_stats_tags,
                    None,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Min", _test_prediction_stats_tags, None),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Max", _test_prediction_stats_tags, None),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Feature/col2/Count", _test_prediction_stats_tags, None),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Mean", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile25", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile50", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Percentile75", _test_prediction_stats_tags, 1),
                (
                    "MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/StandardDeviation",
                    _test_prediction_stats_tags,
                    1,
                ),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Min", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Max", _test_prediction_stats_tags, 1),
                ("MLModel/Sklearn/Named/DummyClassifier/Predict/Label/0/Count", _test_prediction_stats_tags, 1),
            ],
        ),
    ],
)
def test_prediction_stats(run_model, x_train, y_train, x_test, metrics, force_uuid):
    expected_transaction_name = "test_prediction_stats:test_prediction_stats.<locals>._test"

    @validate_transaction_metrics(
        expected_transaction_name,
        dimensional_metrics=metrics,
        background_task=True,
    )
    @background_task()
    def _test():
        run_model(x_train, y_train, x_test)

    _test()


_test_prediction_stats_multilabel_output_tags = frozenset(
    {("modelName", "MultiOutputClassifier"), ("prediction_id", ML_METRIC_FORCED_UUID), ("model_version", "0.0.0")}
)


def test_prediction_stats_multilabel_output(force_uuid):
    expected_transaction_name = "test_prediction_stats:test_prediction_stats_multilabel_output.<locals>._test"
    stats = ["Mean", "Percentile25", "Percentile50", "Percentile75", "StandardDeviation", "Min", "Max", "Count"]
    metrics = [
        (
            "MLModel/Sklearn/Named/MultiOutputClassifier/Predict/Feature/%s/%s" % (feature_col, stat_name),
            _test_prediction_stats_multilabel_output_tags,
            1,
        )
        for feature_col in range(20)
        for stat_name in stats
    ]
    metrics.extend(
        [
            (
                "MLModel/Sklearn/Named/MultiOutputClassifier/Predict/Label/%s/%s" % (label_col, stat_name),
                _test_prediction_stats_multilabel_output_tags,
                1,
            )
            for label_col in range(3)
            for stat_name in stats
        ]
    )

    @validate_transaction_metrics(
        expected_transaction_name,
        dimensional_metrics=metrics,
        background_task=True,
    )
    @background_task()
    def _test():
        from sklearn.datasets import make_multilabel_classification
        from sklearn.linear_model import LogisticRegression
        from sklearn.multioutput import MultiOutputClassifier

        x_train, y_train = make_multilabel_classification(n_classes=3, random_state=0)
        clf = MultiOutputClassifier(LogisticRegression()).fit(x_train, y_train)
        clf.predict([x_train[-1]])

    _test()


@pytest.fixture
def run_model():
    def _run(x_train, y_train, x_test):
        from sklearn import dummy

        clf = dummy.DummyClassifier(random_state=0)
        model = clf.fit(x_train, y_train)

        labels = model.predict(x_test)

    return _run



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from google.protobuf import __version__

    PROTOBUF_VERSION = tuple(int(v) for v in __version__.split("."))
except Exception:
    PROTOBUF_VERSION = (0, 0, 0)

# Import appropriate generated pb2 file for protobuf version
if PROTOBUF_VERSION >= (4,):
    from newrelic.core.infinite_tracing_v4_pb2 import (  # noqa: F401; pylint: disable=W0611
        AttributeValue,
        RecordStatus,
        Span,
        SpanBatch,
    )
else:
    from newrelic.core.infinite_tracing_v3_pb2 import (  # noqa: F401; pylint: disable=W0611
        AttributeValue,
        RecordStatus,
        Span,
        SpanBatch,
    )



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.datastore_trace import datastore_trace
from newrelic.api.external_trace import ExternalTrace
from newrelic.api.message_trace import message_trace
from newrelic.common.object_wrapper import wrap_function_wrapper


def extract_sqs(*args, **kwargs):
    queue_value = kwargs.get("QueueUrl", "Unknown")
    return queue_value.rsplit("/", 1)[-1]


def extract(argument_names, default=None):
    def extractor_list(*args, **kwargs):
        for argument_name in argument_names:
            argument_value = kwargs.get(argument_name)
            if argument_value:
                return argument_value
        return default

    def extractor_string(*args, **kwargs):
        return kwargs.get(argument_names, default)

    if isinstance(argument_names, str):
        return extractor_string

    return extractor_list


CUSTOM_TRACE_POINTS = {
    ("sns", "publish"): message_trace("SNS", "Produce", "Topic", extract(("TopicArn", "TargetArn"), "PhoneNumber")),
    ("dynamodb", "put_item"): datastore_trace("DynamoDB", extract("TableName"), "put_item"),
    ("dynamodb", "get_item"): datastore_trace("DynamoDB", extract("TableName"), "get_item"),
    ("dynamodb", "update_item"): datastore_trace("DynamoDB", extract("TableName"), "update_item"),
    ("dynamodb", "delete_item"): datastore_trace("DynamoDB", extract("TableName"), "delete_item"),
    ("dynamodb", "create_table"): datastore_trace("DynamoDB", extract("TableName"), "create_table"),
    ("dynamodb", "delete_table"): datastore_trace("DynamoDB", extract("TableName"), "delete_table"),
    ("dynamodb", "query"): datastore_trace("DynamoDB", extract("TableName"), "query"),
    ("dynamodb", "scan"): datastore_trace("DynamoDB", extract("TableName"), "scan"),
    ("sqs", "send_message"): message_trace("SQS", "Produce", "Queue", extract_sqs),
    ("sqs", "send_message_batch"): message_trace("SQS", "Produce", "Queue", extract_sqs),
    ("sqs", "receive_message"): message_trace("SQS", "Consume", "Queue", extract_sqs),
}


def bind__create_api_method(py_operation_name, operation_name, service_model, *args, **kwargs):
    return (py_operation_name, service_model)


def _nr_clientcreator__create_api_method_(wrapped, instance, args, kwargs):
    (py_operation_name, service_model) = bind__create_api_method(*args, **kwargs)

    service_name = service_model.service_name.lower()
    tracer = CUSTOM_TRACE_POINTS.get((service_name, py_operation_name))

    wrapped = wrapped(*args, **kwargs)

    if not tracer:
        return wrapped

    return tracer(wrapped)


def _bind_make_request_params(operation_model, request_dict, *args, **kwargs):
    return operation_model, request_dict


def _nr_endpoint_make_request_(wrapped, instance, args, kwargs):
    operation_model, request_dict = _bind_make_request_params(*args, **kwargs)
    url = request_dict.get("url", "")
    method = request_dict.get("method", None)

    with ExternalTrace(library="botocore", url=url, method=method, source=wrapped) as trace:
        try:
            trace._add_agent_attribute("aws.operation", operation_model.name)
        except:
            pass

        result = wrapped(*args, **kwargs)
        try:
            request_id = result[1]["ResponseMetadata"]["RequestId"]
            trace._add_agent_attribute("aws.requestId", request_id)
        except:
            pass
        return result


def instrument_botocore_endpoint(module):
    wrap_function_wrapper(module, "Endpoint.make_request", _nr_endpoint_make_request_)


def instrument_botocore_client(module):
    wrap_function_wrapper(module, "ClientCreator._create_api_method", _nr_clientcreator__create_api_method_)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import threading

from newrelic.api.application import application_instance


def create_request(streaming_request, count=1, timesout=False):
    from sample_application import Message

    def _message_stream():
        for i in range(count):
            yield Message(text="Hello World", count=count, timesout=timesout)

    if streaming_request:
        request = _message_stream()
    else:
        request = Message(text="Hello World", count=count, timesout=timesout)

    return request


def get_result(method, request, *args, **kwargs):
    try:
        from grpc._channel import _InactiveRpcError as Error
    except ImportError:
        from grpc._channel import _Rendezvous as Error
    result = None
    try:
        result = method(request, *args, **kwargs)
        list(result)
    except Error as e:
        result = e
    except Exception:
        pass
    return result


def wait_for_transaction_completion(fn):
    CALLED = threading.Event()
    application = application_instance()
    record_transaction = application.record_transaction

    def record_transaction_wrapper(*args, **kwargs):
        record_transaction(*args, **kwargs)
        CALLED.set()

    @functools.wraps(fn)
    def _waiter(*args, **kwargs):
        application.record_transaction = record_transaction_wrapper
        try:
            result = fn(*args, **kwargs)
            CALLED.wait(timeout=1)
            return result
        finally:
            application.record_transaction = record_transaction

    return _waiter



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

import pytest
import webtest
from testing_support.fixtures import override_application_settings

from newrelic.api.transaction import (
    add_custom_attribute,
    get_browser_timing_footer,
    set_transaction_name,
)
from newrelic.api.wsgi_application import wsgi_application

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
FIXTURE = os.path.join(CURRENT_DIR, "fixtures", "rum_client_config.json")

def _load_tests():
    with open(FIXTURE, "r") as fh:
        js = fh.read()
    return json.loads(js)


fields = [
    "testname",
    "apptime_milliseconds",
    "queuetime_milliseconds",
    "browser_monitoring.attributes.enabled",
    "transaction_name",
    "license_key",
    "connect_reply",
    "user_attributes",
    "expected",
]

# Replace . as not a valid character in python argument names

field_names = ",".join([f.replace(".", "_") for f in fields])


def _parametrize_test(test):
    return tuple([test.get(f, None) for f in fields])


_rum_tests = [_parametrize_test(t) for t in _load_tests()]


@wsgi_application()
def target_wsgi_application(environ, start_response):
    status = "200 OK"

    txn_name = environ.get("txn_name")
    set_transaction_name(txn_name, group="")

    user_attrs = json.loads(environ.get("user_attrs"))
    for key, value in user_attrs.items():
        add_custom_attribute(key, value)

    text = "<html><head>%s</head><body><p>RESPONSE</p></body></html>"

    output = (text % get_browser_timing_footer()).encode("UTF-8")

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]
    start_response(status, response_headers)

    return [output]


target_application = webtest.TestApp(target_wsgi_application)


@pytest.mark.parametrize(field_names, _rum_tests)
def test_browser_montioring(
    testname,
    apptime_milliseconds,
    queuetime_milliseconds,
    browser_monitoring_attributes_enabled,
    transaction_name,
    license_key,
    connect_reply,
    user_attributes,
    expected,
):

    settings = {
        "browser_monitoring.attributes.enabled": browser_monitoring_attributes_enabled,
        "license_key": license_key,
        "js_agent_loader": "<!-- NREUM HEADER -->",
    }
    settings.update(connect_reply)

    @override_application_settings(settings)
    def run_browser_data_test():

        response = target_application.get(
            "/", extra_environ={"txn_name": str(transaction_name), "user_attrs": json.dumps(user_attributes)}
        )

        # We actually put the "footer" in the header, the first script is the
        # agent "header", the second one is where the data lives, hence the [1].

        footer = response.html.html.head.find_all("script")[1]
        footer_data = json.loads(footer.string.split("NREUM.info=")[1])

        # Not feasible to test the time metric values in testing

        expected.pop("queueTime")
        expected.pop("applicationTime")
        assert footer_data["applicationTime"] >= 0
        assert footer_data["queueTime"] >= 0

        # Python always prepends stuff to the transaction name, so this
        # doesn't match the obscured value.

        expected.pop("transactionName")

        # Check that all other values are correct

        for key, value in expected.items():

            # If there are no attributes, the spec allows us to omit the
            # 'atts' field altogether, so we do. But, the cross agent tests
            # don't omit it, so we need to special case 'atts' when we compare
            # to 'expected'.

            if key == "atts" and value == "":
                assert key not in footer_data
            else:
                assert footer_data[key] == value

    run_browser_data_test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from copy import deepcopy

import pytest
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)
from testing_support.validators.validate_transaction_trace_attributes import (
    validate_transaction_trace_attributes,
)

from newrelic.api import lambda_handler


# NOTE: this fixture will force all tests in this file to assume that a cold
#       start has occurred, *except* when a test has a parameter named
#       "is_cold" and its value is True
@pytest.fixture(autouse=True)
def force_cold_start_status(request):
    try:
        is_cold_start = request.getfixturevalue("is_cold")
        lambda_handler.COLD_START_RECORDED = not is_cold_start
    except Exception:
        lambda_handler.COLD_START_RECORDED = True


@lambda_handler.lambda_handler()
def handler(event, context):
    return {
        "statusCode": "200",
        "body": "{}",
        "headers": {
            "Content-Type": "application/json",
            "Content-Length": 2,
        },
    }


_override_settings = {
    "attributes.include": ["request.parameters.*", "request.headers.*"],
}
_expected_attributes = {
    "agent": [
        "aws.requestId",
        "aws.lambda.arn",
        "request.method",
        "request.uri",
        "response.status",
        "response.headers.contentType",
        "response.headers.contentLength",
    ],
    "user": [],
    "intrinsic": [],
}

_exact_attrs = {
    "agent": {
        "request.parameters.foo": "bar",
        "request.headers.host": "myhost",
    },
    "user": {},
    "intrinsic": {},
}

empty_event = {}
firehose_event = {
    "records": [
        {
            "recordId": "495469866831355442",
            "data": "SGVsbG8sIHRoaXMgaXMgYSB0ZXN0IDEyMy4=",
            "approximateArrivalTimestamp": 1495072949453,
        }
    ],
    "region": "us-west-2",
    "deliveryStreamArn": "arn:aws:kinesis:EXAMPLE",
    "invocationId": "invocationIdExample",
}


class Context(object):
    aws_request_id = "cookies"
    invoked_function_arn = "arn"
    function_name = "cats"
    function_version = "$LATEST"
    memory_limit_in_mb = 128


# The lambda_hander has been deprecated for 3+ years
@pytest.mark.skip(reason="The lambda_handler has been deprecated")
@pytest.mark.parametrize("is_cold", (False, True))
def test_lambda_transaction_attributes(is_cold, monkeypatch):
    # setup copies of the attribute lists for this test only
    _forgone_params = {}
    _exact = deepcopy(_exact_attrs)
    _expected = deepcopy(_expected_attributes)

    # if we have a cold start, then we should see aws.lambda.coldStart=True
    if is_cold:
        _exact["agent"]["aws.lambda.coldStart"] = True
        _expected["agent"].append("aws.lambda.coldStart")

    # otherwise, then we need to make sure that we don't see it at all
    else:
        _forgone_params = {"agent": ["aws.lambda.coldStart"], "user": [], "intrinsic": []}

    @validate_transaction_trace_attributes(required_params=_expected, forgone_params=_forgone_params)
    @validate_transaction_event_attributes(
        required_params=_expected, forgone_params=_forgone_params, exact_attrs=_exact
    )
    @override_application_settings(_override_settings)
    def _test():
        monkeypatch.setenv("AWS_REGION", "earth")
        handler(
            {
                "httpMethod": "GET",
                "path": "/",
                "headers": {
                    "HOST": "myhost",
                },
                "queryStringParameters": {"foo": "bar"},
                "multiValueQueryStringParameters": {"foo": ["bar"]},
            },
            Context,
        )

    _test()


# The lambda_hander has been deprecated for 3+ years
@pytest.mark.skip(reason="The lambda_handler has been deprecated")
@validate_transaction_trace_attributes(_expected_attributes)
@validate_transaction_event_attributes(_expected_attributes)
@override_application_settings(_override_settings)
def test_lambda_malformed_api_gateway_payload(monkeypatch):
    monkeypatch.setenv("AWS_REGION", "earth")
    handler(
        {
            "httpMethod": "GET",
            "path": "/",
            "headers": {},
            "queryStringParameters": 42,
            "multiValueQueryStringParameters": 42,
        },
        Context,
    )


_malformed_request_attributes = {
    "agent": [
        "aws.requestId",
        "aws.lambda.arn",
    ],
    "user": [],
    "intrinsic": [],
}


@validate_transaction_trace_attributes(_malformed_request_attributes)
@validate_transaction_event_attributes(_malformed_request_attributes)
@override_application_settings(_override_settings)
def test_lambda_malformed_request_headers():
    handler(
        {
            "httpMethod": "GET",
            "path": "/",
            "headers": None,
        },
        Context,
    )


_malformed_response_attributes = {
    "agent": [
        "aws.requestId",
        "aws.lambda.arn",
        "request.method",
        "request.uri",
        "response.status",
    ],
    "user": [],
    "intrinsic": [],
}


# The lambda_hander has been deprecated for 3+ years
@pytest.mark.skip(reason="The lambda_handler has been deprecated")
@validate_transaction_trace_attributes(_malformed_response_attributes)
@validate_transaction_event_attributes(_malformed_response_attributes)
@override_application_settings(_override_settings)
def test_lambda_malformed_response_headers():
    @lambda_handler.lambda_handler()
    def handler(event, context):
        return {
            "statusCode": 200,
            "body": "{}",
            "headers": None,
        }

    handler(
        {
            "httpMethod": "GET",
            "path": "/",
            "headers": {},
        },
        Context,
    )


_no_status_code_response = {
    "agent": [
        "aws.requestId",
        "aws.lambda.arn",
        "request.method",
        "request.uri",
        "response.headers.contentType",
        "response.headers.contentLength",
    ],
    "user": [],
    "intrinsic": [],
}


# The lambda_hander has been deprecated for 3+ years
@pytest.mark.skip(reason="The lambda_handler has been deprecated")
@validate_transaction_trace_attributes(_no_status_code_response)
@validate_transaction_event_attributes(_no_status_code_response)
@override_application_settings(_override_settings)
def test_lambda_no_status_code_response():
    @lambda_handler.lambda_handler()
    def handler(event, context):
        return {
            "body": "{}",
            "headers": {
                "Content-Type": "application/json",
                "Content-Length": 2,
            },
        }

    handler(
        {
            "httpMethod": "GET",
            "path": "/",
            "headers": {},
        },
        Context,
    )


# The lambda_hander has been deprecated for 3+ years
@pytest.mark.skip(reason="The lambda_handler has been deprecated")
@pytest.mark.parametrize("event,arn", ((empty_event, None), (firehose_event, "arn:aws:kinesis:EXAMPLE")))
def test_lambda_event_source_arn_attribute(event, arn):
    if arn is None:
        _exact = None
        _expected = None
        _forgone = {
            "user": [],
            "intrinsic": [],
            "agent": ["aws.lambda.eventSource.arn"],
        }
    else:
        _exact = {
            "user": {},
            "intrinsic": {},
            "agent": {"aws.lambda.eventSource.arn": arn},
        }
        _expected = {
            "user": [],
            "intrinsic": [],
            "agent": ["aws.lambda.eventSource.arn"],
        }
        _forgone = None

    @validate_transaction_trace_attributes(required_params=_expected, forgone_params=_forgone)
    @validate_transaction_event_attributes(required_params=_expected, forgone_params=_forgone, exact_attrs=_exact)
    @override_application_settings(_override_settings)
    def _test():
        handler(event, Context)

    _test()


# The lambda_hander has been deprecated for 3+ years
@pytest.mark.skip(reason="The lambda_handler has been deprecated")
@pytest.mark.parametrize(
    "api",
    (
        lambda_handler.lambda_handler,
        functools.partial(lambda_handler.LambdaHandlerWrapper, handler),
    ),
)
def test_deprecation_warnings(api):
    with pytest.deprecated_call():
        api()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from graphene import Field, Int, List
from graphene import Mutation as GrapheneMutation
from graphene import NonNull, ObjectType, Schema, String, Union

from ._target_schema_sync import Author, Book, Magazine, Item, Library, Storage, authors, books, magazines, libraries


storage = []


async def resolve_library(self, info, index):
    return libraries[index]

async def resolve_storage(self, info):
    return [storage.pop()]

async def resolve_search(self, info, contains):
    search_books = [b for b in books if contains in b.name]
    search_magazines = [m for m in magazines if contains in m.name]
    return search_books + search_magazines

async def resolve_hello(self, info):
    return "Hello!"

async def resolve_echo(self, info, echo):
    return echo

async def resolve_error(self, info):
    raise RuntimeError("Runtime Error!")

async def resolve_storage_add(self, info, string):
    storage.append(string)
    return StorageAdd(string=string)


class StorageAdd(GrapheneMutation):
    class Arguments:
        string = String(required=True)

    string = String()
    mutate = resolve_storage_add


class Query(ObjectType):
    library = Field(Library, index=Int(required=True), resolver=resolve_library)
    hello = String(resolver=resolve_hello)
    search = Field(List(Item), contains=String(required=True), resolver=resolve_search)
    echo = Field(String, echo=String(required=True), resolver=resolve_echo)
    storage = Field(Storage, resolver=resolve_storage)
    error = String(resolver=resolve_error)
    error_non_null = Field(NonNull(String), resolver=resolve_error)
    error_middleware = String(resolver=resolve_hello)


class Mutation(ObjectType):
    storage_add = StorageAdd.Field()


target_schema = Schema(query=Query, mutation=Mutation, auto_camelcase=False)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pytest

from newrelic.common.utilization import AzureUtilization
from testing_support.mock_http_client import create_client_cls
from testing_support.validators.validate_internal_metrics import validate_internal_metrics


CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
FIXTURE = os.path.normpath(os.path.join(CURRENT_DIR, 'fixtures',
    'utilization_vendor_specific', 'azure.json'))

_parameters_list = ['testname', 'uri', 'expected_vendors_hash',
        'expected_metrics']

_parameters = ','.join(_parameters_list)


def _load_tests():
    with open(FIXTURE, 'r') as fh:
        js = fh.read()
    return json.loads(js)


def _parametrize_test(test):
    return tuple([test.get(f, None) for f in _parameters_list])


_azure_tests = [_parametrize_test(t) for t in _load_tests()]


@pytest.mark.parametrize(_parameters, _azure_tests)
def test_azure(monkeypatch, testname, uri,
               expected_vendors_hash, expected_metrics):

    # Generate mock responses for HttpClient

    def _get_mock_return_value(api_result):
        if api_result['timeout']:
            return 0, None
        else:
            body = json.dumps(api_result['response'])
            return 200, body.encode('utf-8')

    url, api_result = uri.popitem()
    status, data = _get_mock_return_value(api_result)

    client_cls = create_client_cls(status, data, url)
    monkeypatch.setattr(AzureUtilization, "CLIENT_CLS", client_cls)

    metrics = []
    if expected_metrics:
        metrics = [(k, v.get('call_count')) for k, v in
                expected_metrics.items()]

    # Define function that actually runs the test

    @validate_internal_metrics(metrics=metrics)
    def _test_azure_data():

        data = AzureUtilization.detect()

        if data:
            azure_vendor_hash = {'azure': data}
        else:
            azure_vendor_hash = None

        assert azure_vendor_hash == expected_vendors_hash

    _test_azure_data()

    assert not client_cls.FAIL



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import grpc
import pytest
from conftest import create_stub_and_channel
from _test_common import create_request, wait_for_transaction_completion
from newrelic.core.config import global_settings
from testing_support.fixtures import (
        override_application_settings,
        override_generic_settings, function_not_called)
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_transaction_event_attributes import validate_transaction_event_attributes

def select_python_version(py2, py3):
    return six.PY3 and py3 or py2


if hasattr(grpc, '__version__'):
    GRPC_VERSION = tuple(int(v) for v in grpc.__version__.split('.'))
else:
    GRPC_VERSION = None

_test_matrix = ["method_name,streaming_request", [
    ("DoUnaryUnary", False),
    ("DoUnaryStream", False),
    ("DoStreamUnary", True),
    ("DoStreamStream", True)
]]


@pytest.mark.parametrize(*_test_matrix)
def test_simple(method_name, streaming_request, mock_grpc_server, stub):
    port = mock_grpc_server
    request = create_request(streaming_request)
    _transaction_name = \
        "sample_application:SampleApplicationServicer.{}".format(method_name)
    method = getattr(stub, method_name)

    @validate_code_level_metrics("sample_application.SampleApplicationServicer", method_name)
    @validate_transaction_metrics(_transaction_name)
    @override_application_settings({'attributes.include': ['request.*']})
    @validate_transaction_event_attributes(
            required_params={
                'agent': ['request.uri', 'request.headers.userAgent',
                    'response.status', 'response.headers.contentType'],
                'user': [],
                'intrinsic': ['port'],
            },
            exact_attrs={
                'agent': {},
                'user': {},
                'intrinsic': {'port': port}
            })
    @wait_for_transaction_completion
    def _doit():
        response = method(request)

        try:
            list(response)
        except Exception:
            pass

    _doit()


@pytest.mark.parametrize(*_test_matrix)
def test_raises_response_status(method_name, streaming_request,
        mock_grpc_server, stub):
    port = mock_grpc_server
    request = create_request(streaming_request)

    method_name = method_name + 'Raises'

    _transaction_name = \
        "sample_application:SampleApplicationServicer.{}".format(method_name)
    method = getattr(stub, method_name)

    status_code = str(grpc.StatusCode.UNKNOWN.value[0])

    @validate_code_level_metrics("sample_application.SampleApplicationServicer", method_name)
    @validate_transaction_errors(errors=[select_python_version(
        py2='exceptions:AssertionError', py3='builtins:AssertionError')])
    @validate_transaction_metrics(_transaction_name)
    @override_application_settings({'attributes.include': ['request.*']})
    @validate_transaction_event_attributes(
            required_params={
                'agent': ['request.uri', 'request.headers.userAgent',
                    'response.status'],
                'user': [],
                'intrinsic': ['port'],
            },
            exact_attrs={
                'agent': {'response.status': status_code},
                'user': {},
                'intrinsic': {'port': port}
            })
    @wait_for_transaction_completion
    def _doit():
        try:
            response = method(request)
            list(response)
        except Exception:
            pass

    _doit()


@pytest.mark.parametrize(*_test_matrix)
def test_abort(method_name, streaming_request, mock_grpc_server, stub):
    method_name += 'Abort'
    port = mock_grpc_server
    request = create_request(streaming_request)
    method = getattr(stub, method_name)

    @validate_code_level_metrics("sample_application.SampleApplicationServicer", method_name)
    @validate_transaction_errors(errors=[select_python_version(
        py2='exceptions:Exception', py3='builtins:Exception')])
    @wait_for_transaction_completion
    def _doit():
        with pytest.raises(grpc.RpcError) as error:
            response = method(request)
            list(response)

        assert error.value.details() == 'aborting'
        assert error.value.code() == grpc.StatusCode.ABORTED

    _doit()


@pytest.mark.parametrize(*_test_matrix)
def test_abort_with_status(method_name, streaming_request, mock_grpc_server, stub):
    method_name += 'AbortWithStatus'
    port = mock_grpc_server
    request = create_request(streaming_request)
    method = getattr(stub, method_name)

    @validate_code_level_metrics("sample_application.SampleApplicationServicer", method_name)
    @validate_transaction_errors(errors=[select_python_version(
        py2='exceptions:Exception', py3='builtins:Exception')])
    @wait_for_transaction_completion
    def _doit():
        with pytest.raises(grpc.RpcError) as error:
            response = method(request)
            list(response)

        assert error.value.details() == 'abort_with_status'
        assert error.value.code() == grpc.StatusCode.ABORTED

    _doit()


def test_no_exception_client_close(mock_grpc_server):
    port = mock_grpc_server
    # We can't use the stub_and_channel fixture here as closing 
    # that channel will cause any subsequent tests to fail.
    # Instead we create a brand new channel to close.
    stub, channel = create_stub_and_channel(port)

    with channel:
        request = create_request(False, timesout=True)

        method = getattr(stub, 'DoUnaryUnary')

        @validate_transaction_errors(errors=[])
        @wait_for_transaction_completion
        def _doit():
            future_response = method.future(request)
            channel.close()
            with pytest.raises(grpc.RpcError) as error:
                future_response.result()

            assert error.value.code() == grpc.StatusCode.CANCELLED

        _doit()


def test_newrelic_disabled_no_transaction(mock_grpc_server, stub):
    port = mock_grpc_server
    request = create_request(False)

    method = getattr(stub, 'DoUnaryUnary')

    @override_generic_settings(global_settings(), {'enabled': False})
    @function_not_called('newrelic.core.stats_engine',
        'StatsEngine.record_transaction')
    @wait_for_transaction_completion
    def _doit():
        method(request)

    _doit()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from newrelic.api.background_task import background_task
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics


@pytest.fixture(scope="session")
def feedparser():
    import feedparser
    return feedparser


@pytest.mark.parametrize("url", (
    "http://localhost",
    "feed:http://localhost",
    "feed://localhost",
))
def test_feedparser_external(feedparser, server, url):
    url = url + ':' + str(server.port)

    @validate_transaction_metrics(
        "test_feedparser_external",
        background_task=True,
        scoped_metrics=(("External/localhost:%d/feedparser/GET" % server.port, 1),),
    )
    @background_task(name="test_feedparser_external")
    def _test():
        feed = feedparser.parse(url)
        assert feed["feed"]["link"] == u"https://pypi.org/"

    _test()


@pytest.mark.parametrize("stream", (True, False))
def test_feedparser_file(feedparser, stream, server):

    @validate_transaction_metrics(
        "test_feedparser_file",
        background_task=True,
        scoped_metrics=(("External/localhost:%d/feedparser/GET" % server.port, None),),
    )
    @background_task(name="test_feedparser_file")
    def _test():
        if stream:
            with open("packages.xml", "rb") as f:
                feed = feedparser.parse(f)
        else:
            feed = feedparser.parse("packages.xml")
        assert feed["feed"]["link"] == u"https://pypi.org/"

    _test()


@pytest.mark.parametrize("url", (
    "http://localhost",
    "packages.xml",
))
def test_feedparser_no_transaction(feedparser, server, url):
    if url.startswith('http://'):
        url = url + ':' + str(server.port)
    feed = feedparser.parse(url)
    assert feed["feed"]["link"] == u"https://pypi.org/"



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket

import cheroot.wsgi
import newrelic.api.transaction
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics


def get_open_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    port = s.getsockname()[1]
    s.close()
    return port


def wsgi_test_app(environ, start_response):
    newrelic.api.transaction.set_transaction_name('wsgi_test_transaction')
    status = '200 OK'
    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)
    return [b'Hello world!']


def start_response(status, response_headers, exc_info=None):
    """Empty callable"""
    pass


_test_scoped_metrics = [
    ('Python/WSGI/Finalize', 1),
    ('Python/WSGI/Application', 1),
    ('Function/test_wsgi:wsgi_test_app', 1),
]


@validate_transaction_metrics('wsgi_test_transaction',
        scoped_metrics=_test_scoped_metrics)
def test_wsgi_test_function_transaction_metrics_positional_args():
    server = cheroot.wsgi.Server(('127.0.0.1', get_open_port()), wsgi_test_app)
    environ = {'REQUEST_URI': '/'}
    resp = server.wsgi_app(environ, start_response)

    if hasattr(resp, 'close'):
        resp.close()


@validate_transaction_metrics('wsgi_test_transaction',
        scoped_metrics=_test_scoped_metrics)
def test_wsgi_test_function_transaction_metrics_keyword_args():
    server = cheroot.wsgi.Server(bind_addr=('127.0.0.1', get_open_port()),
                                 wsgi_app=wsgi_test_app)
    environ = {'REQUEST_URI': '/'}
    resp = server.wsgi_app(environ, start_response)

    if hasattr(resp, 'close'):
        resp.close()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testing_support.validators.validate_datastore_trace_inputs import (
    validate_datastore_trace_inputs,
)

from newrelic.api.background_task import background_task
from newrelic.api.datastore_trace import DatastoreTrace, DatastoreTraceWrapper


@validate_datastore_trace_inputs(
    operation="test_operation",
    target="test_target",
    host="test_host",
    port_path_or_id="test_port",
    database_name="test_db_name",
)
@background_task()
def test_dt_trace_all_args():
    with DatastoreTrace(
        product="Agent Features",
        target="test_target",
        operation="test_operation",
        host="test_host",
        port_path_or_id="test_port",
        database_name="test_db_name",
    ):
        pass


@validate_datastore_trace_inputs(operation=None, target=None, host=None, port_path_or_id=None, database_name=None)
@background_task()
def test_dt_trace_empty():
    with DatastoreTrace(product=None, target=None, operation=None):
        pass


@background_task()
def test_dt_trace_callable_args():
    def product_callable():
        return "Agent Features"

    def target_callable():
        return "test_target"

    def operation_callable():
        return "test_operation"

    def host_callable():
        return "test_host"

    def port_path_id_callable():
        return "test_port"

    def db_name_callable():
        return "test_db_name"

    @validate_datastore_trace_inputs(
        operation="test_operation",
        target="test_target",
        host="test_host",
        port_path_or_id="test_port",
        database_name="test_db_name",
    )
    def _test():
        pass

    wrapped_fn = DatastoreTraceWrapper(
        _test,
        product=product_callable,
        target=target_callable,
        operation=operation_callable,
        host=host_callable,
        port_path_or_id=port_path_id_callable,
        database_name=db_name_callable,
    )
    wrapped_fn()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import webtest

from newrelic.packages import six

from testing_support.validators.validate_transaction_errors import validate_transaction_errors

import cherrypy

is_ge_cherrypy32 = (tuple(map(int,
        cherrypy.__version__.split('.')[:2])) >= (3,2))

requires_cherrypy32 = pytest.mark.skipif(not is_ge_cherrypy32,
        reason="The dispatch mechanism was only added in CherryPy 3.2.")

class Resource(object):

    def _cp_dispatch(self, vpath):
        raise RuntimeError('dispatch error')

if is_ge_cherrypy32:
    dispatcher = cherrypy.dispatch.MethodDispatcher()

    conf = { '/': { 'request.dispatch': dispatcher } }

    application = cherrypy.Application(Resource(), '/', conf)
    test_application = webtest.TestApp(application)

if six.PY3:
    _test_dispatch_exception_errors = ['builtins:RuntimeError']
else:
    _test_dispatch_exception_errors = ['exceptions:RuntimeError']

@requires_cherrypy32
@validate_transaction_errors(errors=_test_dispatch_exception_errors)
def test_dispatch_exception():
    response = test_application.get('/sub/a/b', status=500)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ._target_schema_async import target_schema as target_schema_async
from ._target_schema_sync import target_schema as target_schema_sync
from framework_graphene.test_application import GRAPHENE_VERSION


def check_response(query, response):
    if isinstance(query, str) and "error" not in query:
        assert not response.errors, response
        assert response.data
    else:
        assert response.errors, response


def run_sync(schema):
    def _run_sync(query, middleware=None):
        response = schema.execute(query, middleware=middleware)
        check_response(query, response)

        return response.data

    return _run_sync


def run_async(schema):
    import asyncio

    def _run_async(query, middleware=None):
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(schema.execute_async(query, middleware=middleware))
        check_response(query, response)

        return response.data

    return _run_async

target_application = {
    "sync-sync": run_sync(target_schema_sync),
    "async-sync": run_async(target_schema_sync),
    "async-async": run_async(target_schema_async),
    }





# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webtest

from pyramid.response import Response
from pyramid.view import view_config, notfound_view_config
from pyramid.config import Configurator
import pyramid.httpexceptions as exc


@view_config(route_name='home')
def home_view(request):
    return Response('<p>INDEX RESPONSE</p>')

@notfound_view_config(append_slash=True)
def not_found(request):
    return Response('<p>NOT FOUND</p>', status='404 Not Found')

config = Configurator()
config.add_route('home', '/')
config.scan()

application = config.make_wsgi_app()

_test_application = webtest.TestApp(application)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
import pytest
import newrelic.packages.six as six

from newrelic.api.background_task import BackgroundTask
from newrelic.api.application import application_instance
from newrelic.common.object_wrapper import transient_function_wrapper, function_wrapper


@function_wrapper
def capture_errors(wrapped, instance, args, kwargs):
    ERRORS = []

    @transient_function_wrapper(
            'newrelic.api.transaction', 'Transaction.__exit__')
    def capture_errors(wrapped, instance, args, kwargs):
        try:
            return wrapped(*args, **kwargs)
        except Exception as e:
            ERRORS.append(e)
            raise

    result = capture_errors(wrapped)(*args, **kwargs)

    assert not ERRORS
    return result


@pytest.mark.parametrize('circular', (True, False))
@capture_errors
def test_dead_transaction_ends(circular):
    if circular and six.PY2:
        pytest.skip("Circular references in py2 result in a memory leak. "
                "There is no way to remove transactions from the weakref "
                "cache in this case.")

    transaction = BackgroundTask(
            application_instance(), "test_dead_transaction_ends")
    if circular:
        transaction._self = transaction

    transaction.__enter__()
    del transaction
    gc.collect()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# from conftest import AIOREDIS_VERSION, event_loop, loop
from testing_support.db_settings import redis_settings
from testing_support.fixtures import override_application_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task

DB_SETTINGS = redis_settings()[0]


_enable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
}

_disable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
}


_base_scoped_metrics = (
    ("Datastore/operation/Redis/get", 1),
    ("Datastore/operation/Redis/set", 1),
)

_base_rollup_metrics = (
    ("Datastore/all", 2),
    ("Datastore/allOther", 2),
    ("Datastore/Redis/all", 2),
    ("Datastore/Redis/allOther", 2),
    ("Datastore/operation/Redis/get", 1),
    ("Datastore/operation/Redis/set", 1),
)

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

_host = instance_hostname(DB_SETTINGS["host"])
_port = DB_SETTINGS["port"]

_instance_metric_name = "Datastore/instance/Redis/%s/%s" % (_host, _port)

_enable_rollup_metrics.append((_instance_metric_name, 2))

_disable_rollup_metrics.append((_instance_metric_name, None))


async def exercise_redis(client, key):
    await client.set(key, "value")
    await client.get(key)


@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
    "test_get_and_set:test_redis_client_operation_enable_instance",
    scoped_metrics=_enable_scoped_metrics,
    rollup_metrics=_enable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_redis_client_operation_enable_instance(client, loop, key):
    loop.run_until_complete(exercise_redis(client, key))


@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
    "test_get_and_set:test_redis_client_operation_disable_instance",
    scoped_metrics=_disable_scoped_metrics,
    rollup_metrics=_disable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_redis_client_operation_disable_instance(client, loop, key):
    loop.run_until_complete(exercise_redis(client, key))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from newrelic.api.transaction import current_transaction
from newrelic.api.background_task import background_task

from testing_support.db_settings import redis_settings
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.util import instance_hostname


DB_SETTINGS = redis_settings()[0]

_enable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
    "datastore_tracer.database_name_reporting.enabled": True,
    "distributed_tracing.enabled": True,
    "span_events.enabled": True,
}
_disable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
    "datastore_tracer.database_name_reporting.enabled": False,
    "distributed_tracing.enabled": True,
    "span_events.enabled": True,
}

async def _exercise_db(client):
    await client.set("key", "value")
    await client.get("key")

    if hasattr(client, "execute_command"):
        await client.execute_command("CLIENT", "LIST", parse="LIST")
    else:
        await client.execute("CLIENT", "LIST")


@pytest.mark.parametrize("db_instance_enabled", (True, False))
@pytest.mark.parametrize("instance_enabled", (True, False))
def test_span_events(client, instance_enabled, db_instance_enabled, loop):
    guid = "dbb533c53b749e0b"
    priority = 0.5

    common = {
        "type": "Span",
        "transactionId": guid,
        "priority": priority,
        "sampled": True,
        "category": "datastore",
        "component": "Redis",
        "span.kind": "client",
    }
    exact_agents = {}

    if instance_enabled:
        settings = _enable_instance_settings.copy()
        hostname = instance_hostname(DB_SETTINGS["host"])
        exact_agents.update(
            {
                "peer.address": "%s:%s" % (hostname, DB_SETTINGS["port"]),
                "peer.hostname": hostname,
            }
        )
    else:
        settings = _disable_instance_settings.copy()
        exact_agents.update(
            {
                "peer.address": "Unknown:Unknown",
                "peer.hostname": "Unknown",
            }
        )

    if db_instance_enabled and instance_enabled:
        exact_agents.update(
            {
                "db.instance": "0",
            }
        )
        unexpected_agents = ()
    else:
        settings["attributes.exclude"] = ["db.instance"]
        unexpected_agents = ("db.instance",)

    query_1 = common.copy()
    query_1["name"] = "Datastore/operation/Redis/set"

    query_2 = common.copy()
    query_2["name"] = "Datastore/operation/Redis/get"

    query_3 = common.copy()
    query_3["name"] = "Datastore/operation/Redis/client_list"

    @validate_span_events(
        count=1,
        exact_intrinsics=query_1,
        unexpected_intrinsics=("db.instance"),
        exact_agents=exact_agents,
        unexpected_agents=unexpected_agents,
    )
    @validate_span_events(
        count=1,
        exact_intrinsics=query_2,
        unexpected_intrinsics=("db.instance"),
        exact_agents=exact_agents,
        unexpected_agents=unexpected_agents,
    )
    @validate_span_events(
        count=1,
        exact_intrinsics=query_3,
        unexpected_intrinsics=("db.instance"),
        exact_agents=exact_agents,
        unexpected_agents=unexpected_agents,
    )
    @override_application_settings(settings)
    @background_task(name="span_events")
    def _test():
        txn = current_transaction()
        txn.guid = guid
        txn._priority = priority
        txn._sampled = True
        loop.run_until_complete(_exercise_db(client))

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pytest

from newrelic.api.application import application_instance
from newrelic.api.background_task import background_task
from newrelic.api.transaction import record_custom_metric
from newrelic.core.rules_engine import RulesEngine

from testing_support.validators.validate_metric_payload import validate_metric_payload

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
FIXTURE = os.path.normpath(os.path.join(
        CURRENT_DIR, 'fixtures', 'rules.json'))


def _load_tests():
    with open(FIXTURE, 'r') as fh:
        js = fh.read()
    return json.loads(js)


def _make_case_insensitive(rules):
    # lowercase each rule
    for rule in rules:
        rule['match_expression'] = rule['match_expression'].lower()
        if rule.get('replacement'):
            rule['replacement'] = rule['replacement'].lower()
    return rules


@pytest.mark.parametrize('test_group', _load_tests())
def test_rules_engine(test_group):

    # FIXME: The test fixture assumes that matching is case insensitive when it
    # is not. To avoid errors, just lowercase all rules, inputs, and expected
    # values.
    test_rules = _make_case_insensitive(test_group['rules'])
    rules_engine = RulesEngine(test_rules)

    for test in test_group['tests']:

        # lowercase each value
        input_str = test['input'].lower()
        expected = (test['expected'] or '').lower()

        result, ignored = rules_engine.normalize(input_str)

        # When a transaction is to be ignored, the test fixture expects that
        # "expected" is None.
        if ignored:
            assert expected == ''
        else:
            assert result == expected


@pytest.mark.parametrize('test_group', _load_tests())
def test_rules_engine_metric_harvest(test_group):
    # FIXME: The test fixture assumes that matching is case insensitive when it
    # is not. To avoid errors, just lowercase all rules, inputs, and expected
    # values.
    test_rules = _make_case_insensitive(test_group['rules'])
    rules_engine = RulesEngine(test_rules)

    # Set rules engine on core application
    api_application = application_instance(activate=False)
    api_name = api_application.name
    core_application = api_application._agent.application(api_name)
    old_rules = core_application._rules_engine["metric"]  # save previoius rules
    core_application._rules_engine["metric"] = rules_engine

    def send_metrics():
        # Send all metrics in this test batch in one transaction, then harvest so the normalizer is run.
        @background_task(name="send_metrics")
        def _test():
            for test in test_group['tests']:
                # lowercase each value
                input_str = test['input'].lower()
                record_custom_metric(input_str, {"count": 1})
        _test()
        core_application.harvest()

    try:
        # Create a map of all result metrics to validate after harvest
        test_metrics = []
        for test in test_group['tests']:
            expected = (test['expected'] or '').lower()
            if expected == '':  # Ignored
                test_metrics.append((expected, None))
            else:
                test_metrics.append((expected, 1))
        
        # Harvest and validate resulting payload
        validate_metric_payload(metrics=test_metrics)(send_metrics)()
    finally:
        # Replace original rules engine
        core_application._rules_engine["metric"] = old_rules



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import mock
import os
import pytest

import newrelic.common.utilization as u

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
DOCKER_FIXTURE = os.path.join(CURRENT_DIR, 'fixtures', 'docker_container_id')


def _load_docker_test_attributes():
    """Returns a list of docker test attributes in the form:
       [(<filename>, <containerId>), ...]

    """
    docker_test_attributes = []
    test_cases = os.path.join(DOCKER_FIXTURE, 'cases.json')
    with open(test_cases, 'r') as fh:
        js = fh.read()
    json_list = json.loads(js)
    for json_record in json_list:
        docker_test_attributes.append(
            (json_record['filename'], json_record['containerId']))
    return docker_test_attributes


def mock_open(mock_file):
    def _mock_open(filename, mode):
        if filename == "/proc/self/mountinfo":
            raise FileNotFoundError()
        elif filename == "/proc/self/cgroup":
            return mock_file
        raise RuntimeError()
    return _mock_open


@pytest.mark.parametrize('filename, containerId',
                          _load_docker_test_attributes())
def test_docker_container_id_v1(monkeypatch, filename, containerId):
    path = os.path.join(DOCKER_FIXTURE, filename)
    with open(path, 'rb') as f:
        monkeypatch.setattr(u, "open", mock_open(f), raising=False)
        if containerId is not None:
            assert u.DockerUtilization.detect() == {'id': containerId}
        else:
            assert u.DockerUtilization.detect() is None



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pytest
import sys
import tempfile

from newrelic.common.system_info import BootIdUtilization

from testing_support.validators.validate_internal_metrics import validate_internal_metrics


CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
SYS_PLATFORM = sys.platform
FIXTURE = os.path.normpath(os.path.join(CURRENT_DIR, 'fixtures', 'utilization',
    'boot_id.json'))

_parameters_list = ['testname', 'input_total_ram_mib',
        'input_logical_processors', 'input_hostname', 'input_boot_id',
        'expected_output_json', 'expected_metrics']

_parameters = ','.join(_parameters_list)


def _load_tests():
    with open(FIXTURE, 'r') as fh:
        js = fh.read()
    return json.loads(js)


def _parametrize_test(test):
    return tuple([test.get(f, None) for f in _parameters_list])


_boot_id_tests = [_parametrize_test(t) for t in _load_tests()]


class MockedBootIdEndpoint(object):
    def __init__(self, boot_id):
        self.boot_id = boot_id

    def __enter__(self):
        if self.boot_id is not None:
            self.boot_id_file = tempfile.NamedTemporaryFile()
            self.boot_id_file.write(self.boot_id.encode('utf8'))
            self.boot_id_file.seek(0)
            BootIdUtilization.METADATA_URL = self.boot_id_file.name
        else:
            BootIdUtilization.METADATA_URL = '/file/does/not/exist/I/hope'
        sys.platform = 'linux-mock-testing'  # ensure boot_id is gathered

    def __exit__(self, *args, **kwargs):
        sys.platform = SYS_PLATFORM
        if self.boot_id:
            del self.boot_id_file  # close and thus delete the tempfile


@pytest.mark.parametrize(_parameters, _boot_id_tests)
def test_boot_id(testname, input_total_ram_mib, input_logical_processors,
        input_hostname, input_boot_id, expected_output_json, expected_metrics):

    metrics = []
    if expected_metrics:
        metrics = [(k, v.get('call_count')) for k, v in
                expected_metrics.items()]

    @validate_internal_metrics(metrics=metrics)
    def _test_boot_id_data():
        data = BootIdUtilization.detect()

        assert data == expected_output_json.get('boot_id')

    with MockedBootIdEndpoint(input_boot_id):
        _test_boot_id_data()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from newrelic.common.object_names import callable_name
from newrelic.hooks.mlmodel_sklearn import _nr_instrument_model


def wrap_mlmodel(model, name=None, version=None, feature_names=None, label_names=None, metadata=None):
    model_callable_name = callable_name(model)
    _class = model.__class__.__name__
    module = sys.modules[model_callable_name.split(":")[0]]
    _nr_instrument_model(module, _class)
    if name:
        model._nr_wrapped_name = name
    if version:
        model._nr_wrapped_version = version
    if feature_names:
        model._nr_wrapped_feature_names = feature_names
    if label_names:
        model._nr_wrapped_label_names = label_names
    if metadata:
        model._nr_wrapped_metadata = metadata



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from testing_support.fixtures import collector_agent_registration_fixture, collector_available_fixture  # noqa: F401; pylint: disable=W0611
from testing_support.fixtures import (  # noqa: F401; pylint: disable=W0611
    newrelic_caplog as caplog,
)

from newrelic.packages import six


_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
    "debug.log_autorum_middleware": True,
    "agent_limits.errors_per_harvest": 100,
    "ml_insights_events.enabled": True
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (agent_features)", default_settings=_default_settings
)


if six.PY2:
    collect_ignore = [
        "test_async_context_propagation.py",
        "test_coroutine_trace.py",
        "test_coroutine_transaction.py",
        "test_async_timing.py",
        "test_event_loop_wait_time.py",
        "test_asgi_transaction.py",
        "test_asgi_browser.py",
        "test_asgi_distributed_tracing.py",
        "test_asgi_w3c_trace_context.py",
    ]
else:
    from testing_support.fixture.event_loop import event_loop



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from framework_graphql.test_application import *
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_count import (
    validate_transaction_count,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version

STRAWBERRY_VERSION = get_package_version("strawberry-graphql")


@pytest.fixture(scope="session", params=["sync-sync", "async-sync", "async-async", "asgi-sync", "asgi-async"])
def target_application(request):
    from ._target_application import target_application

    target_application = target_application[request.param]

    is_asgi = "asgi" in request.param
    schema_type = request.param.split("-")[1]

    assert STRAWBERRY_VERSION is not None
    return "Strawberry", STRAWBERRY_VERSION, target_application, not is_asgi, schema_type, 0


@pytest.mark.parametrize("capture_introspection_setting", (True, False))
def test_introspection_transactions(target_application, capture_introspection_setting):
    framework, version, target_application, is_bg, schema_type, extra_spans = target_application

    txn_ct = 1 if capture_introspection_setting else 0

    @override_application_settings(
        {"instrumentation.graphql.capture_introspection_queries": capture_introspection_setting}
    )
    @validate_transaction_count(txn_ct)
    @background_task()
    def _test():
        response = target_application("{ __schema { types { name } } }")

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import webtest

from newrelic.packages import six

from testing_support.fixtures import (
        override_application_settings, 
        override_ignore_status_codes)
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics
from testing_support.validators.validate_transaction_errors import validate_transaction_errors

import cherrypy

CHERRYPY_VERSION = tuple(int(v) for v in cherrypy.__version__.split('.'))


class Application(object):

    @cherrypy.expose
    def index(self):
        return 'INDEX RESPONSE'

    @cherrypy.expose
    def error(self):
        raise RuntimeError('error')

    @cherrypy.expose
    def not_found(self):
        raise cherrypy.NotFound()

    @cherrypy.expose
    def not_found_as_http_error(self):
        raise cherrypy.HTTPError(404)

    @cherrypy.expose
    def not_found_as_str_http_error(self):
        raise cherrypy.HTTPError('404 Not Found')

    @cherrypy.expose
    def bad_http_error(self):
        # this will raise HTTPError with status code 500 because 10 is not a
        # valid status code
        raise cherrypy.HTTPError('10 Invalid status code')

    @cherrypy.expose
    def internal_redirect(self):
        raise cherrypy.InternalRedirect('/')

    @cherrypy.expose
    def external_redirect(self):
        raise cherrypy.HTTPRedirect('/')

    @cherrypy.expose
    def upload_files(self, files):
        return 'UPLOAD FILES RESPONSE'

    @cherrypy.expose
    def encode_multipart(self, field, files):
        return 'ENCODE MULTIPART RESPONSE'

    @cherrypy.expose
    def html_insertion(self):
        return ('<!DOCTYPE html><html><head>Some header</head>'
            '<body><h1>My First Heading</h1><p>My first paragraph.</p>'
            '</body></html>')


application = cherrypy.Application(Application())
test_application = webtest.TestApp(application)


@validate_code_level_metrics("test_application.Application", "index")
@validate_transaction_errors(errors=[])
def test_application_index():
    response = test_application.get('')
    response.mustcontain('INDEX RESPONSE')


@validate_transaction_errors(errors=[])
def test_application_index_agent_disabled():
    environ = {'newrelic.enabled': False}
    response = test_application.get('', extra_environ=environ)
    response.mustcontain('INDEX RESPONSE')


@validate_transaction_errors(errors=[])
def test_application_missing():
    test_application.get('/missing', status=404)


if six.PY3:
    _test_application_unexpected_exception_errors = ['builtins:RuntimeError']
else:
    _test_application_unexpected_exception_errors = ['exceptions:RuntimeError']


@validate_transaction_errors(
        errors=_test_application_unexpected_exception_errors)
def test_application_unexpected_exception():
    test_application.get('/error', status=500)


@validate_transaction_errors(errors=[])
def test_application_not_found():
    test_application.get('/not_found', status=404)


@validate_transaction_errors(errors=[])
def test_application_not_found_as_http_error():
    test_application.get('/not_found_as_http_error', status=404)


@validate_transaction_errors(errors=[])
def test_application_internal_redirect():
    response = test_application.get('/internal_redirect')
    response.mustcontain('INDEX RESPONSE')


@validate_transaction_errors(errors=[])
def test_application_external_redirect():
    test_application.get('/external_redirect', status=302)


@validate_transaction_errors(errors=[])
def test_application_upload_files():
    test_application.post('/upload_files', upload_files=[('files', __file__)])


@validate_transaction_errors(errors=[])
def test_application_encode_multipart():
    content_type, body = test_application.encode_multipart(
            params=[('field', 'value')], files=[('files', __file__)])
    test_application.request('/encode_multipart', method='POST',
            content_type=content_type, body=body)


_test_html_insertion_settings = {
    'browser_monitoring.enabled': True,
    'browser_monitoring.auto_instrument': True,
    'js_agent_loader': u'<!-- NREUM HEADER -->',
}


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion():
    response = test_application.get('/html_insertion')

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain('NREUM HEADER', 'NREUM.info')


_error_endpoints = ['/not_found_as_http_error']
if CHERRYPY_VERSION >= (3, 2):
    _error_endpoints.extend(['/not_found_as_str_http_error',
        '/bad_http_error'])


@pytest.mark.parametrize('endpoint', _error_endpoints)
@pytest.mark.parametrize('ignore_overrides,expected_errors', [
    ([], ['cherrypy._cperror:HTTPError']),
    ([404, 500], []),
])
def test_ignore_status_code(endpoint, ignore_overrides, expected_errors):

    @validate_transaction_errors(errors=expected_errors)
    @override_ignore_status_codes(ignore_overrides)
    def _test():
        test_application.get(endpoint, status=[404, 500])

    _test()


@validate_transaction_errors(errors=[])
def test_ignore_status_unexpected_param():
    response = test_application.get('/?arg=1', status=404)
    response.mustcontain(no=['INDEX RESPONSE'])
